from inmanta.plugins import plugin


@plugin
def test_nullable(param: "string?") -> "bool":
    return True


@plugin
def test_not_nullable(param: "string") -> "bool":
    return True


@plugin
def test_nullable_list(param: "number[]?") -> "bool":
    return True


@plugin
def test_not_nullable_list(param: "number[]") -> "bool":
    return True


@plugin
def test_returns_none() -> "string?":
    return None



"""
    Copyright 2017 Inmanta

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Contact: code@inmanta.com
"""

from abc import abstractmethod
from typing import Deque, Dict, Generic, Hashable, List, Optional, Set, TypeVar, Union, cast

import inmanta.warnings as inmanta_warnings
from inmanta.ast import (
    AttributeException,
    CompilerDeprecationWarning,
    CompilerException,
    DoubleSetException,
    Locatable,
    Location,
    ModifiedAfterFreezeException,
    Namespace,
    NotFoundException,
    OptionalValueException,
    RuntimeException,
)
from inmanta.ast.type import Type
from inmanta.execute import dataflow, proxy
from inmanta.execute.dataflow import DataflowGraph
from inmanta.execute.tracking import Tracker
from inmanta.execute.util import Unknown

try:
    from typing import TYPE_CHECKING
except ImportError:
    TYPE_CHECKING = False

if TYPE_CHECKING:
    from inmanta.ast.attribute import Attribute, RelationAttribute
    from inmanta.ast.blocks import BasicBlock
    from inmanta.ast.entity import Default, Entity, EntityLike, Implement, Implementation  # noqa: F401
    from inmanta.ast.statements import ExpressionStatement, RawResumer, Resumer, Statement
    from inmanta.compiler import Compiler


T = TypeVar("T")


class ResultCollector(Generic[T]):
    """
    Helper interface for gradual execution
    """

    __slots__ = ()

    def receive_result(self, value: T, location: Location) -> None:
        """
        receive a possibly partial result
        """
        raise NotImplementedError()


class IPromise(Generic[T]):
    __slots__ = ()

    @abstractmethod
    def set_value(self, value: T, location: Location, recur: bool = True) -> None:
        pass


class ResultVariable(ResultCollector[T], IPromise[T]):
    """
    A ResultVariable is like a future
     - it has a list of waiters
     - when a value is set, the waiters are notified,
        they decrease their wait count and
        queue themselves when their wait count becomes 0

    If a type is set on a result variable, setting a value of another type will produce an exception.

    In order to assist heuristic evaluation, result variables keep track of any statement that will assign a value to it
    """

    location: Location

    __slots__ = ("location", "provider", "waiters", "value", "hasValue", "type", "_node")

    def __init__(self, value: Optional[T] = None) -> None:
        self.provider: "Optional[Statement]" = None
        self.waiters: "List[Waiter]" = []
        self.value: Optional[T] = value
        self.hasValue: bool = False
        self.type: Optional[Type] = None
        self._node: Optional[dataflow.AssignableNodeReference] = None

    def set_type(self, mytype: Type) -> None:
        self.type = mytype

    def set_provider(self, provider: "Statement") -> None:
        # no checking for double set, this is done in the actual assignment
        self.provider = provider

    def get_promise(self, provider: "Statement") -> IPromise[T]:
        """Alternative for set_provider for better handling of ListVariables."""
        self.provider = provider
        return self

    def is_ready(self) -> bool:
        return self.hasValue

    def waitfor(self, waiter: "Waiter") -> None:
        if self.is_ready():
            waiter.ready(self)
        else:
            self.waiters.append(waiter)

    def set_value(self, value: T, location: Location, recur: bool = True) -> None:
        if self.hasValue:
            if self.value != value:
                raise DoubleSetException(self, None, value, location)
            else:
                return
        if not isinstance(value, Unknown) and self.type is not None:
            self.type.validate(value)
        self.value = value
        self.location = location
        self.hasValue = True
        for waiter in self.waiters:
            waiter.ready(self)
        # prevent memory leaks
        self.waiters = None

    def get_value(self) -> T:
        if not self.hasValue:
            raise proxy.UnsetException("Value not available", self)

        return self.value

    def can_get(self) -> bool:
        return self.hasValue

    def freeze(self) -> None:
        pass

    def receive_result(self, value: T, location: Location) -> None:
        pass

    def listener(self, resulcollector: ResultCollector[T], location: Location) -> None:
        """
        add a listener to report new values to, only for lists
        """
        pass

    def is_multi(self) -> bool:
        return False

    def set_dataflow_node(self, node: dataflow.AssignableNodeReference) -> None:
        assert self._node is None or self._node == node
        self._node = node
        self._node.set_result_variable(self)

    def get_dataflow_node(self) -> dataflow.AssignableNodeReference:
        assert self._node is not None, "assertion error at %s.get_dataflow_node() in ResultVariable" % self
        return self._node


class AttributeVariable(ResultVariable["Instance"]):
    """
    a result variable for a relation with arity 1

    when assigned a value, it will also assign a value to its inverse relation
    """

    __slots__ = ("attribute", "myself")

    def __init__(self, attribute: "RelationAttribute", instance: "Instance"):
        self.attribute: "RelationAttribute" = attribute
        self.myself: "Instance" = instance
        ResultVariable.__init__(self)

    def set_value(self, value: "Instance", location: Location, recur: bool = True) -> None:
        if self.hasValue:
            if self.value != value:
                raise DoubleSetException(self, None, value, location)
            else:
                return
        if not isinstance(value, Unknown) and self.type is not None:
            self.type.validate(value)
        self.value = value
        self.location = location
        self.hasValue = True
        # set counterpart
        if self.attribute.end and recur:
            assert isinstance(value, Instance)
            value.set_attribute(self.attribute.end.name, self.myself, location, False)
        for waiter in self.waiters:
            waiter.ready(self)
        self.waiters = None


class DelayedResultVariable(ResultVariable[T]):
    """
    DelayedResultVariable are ResultVariables of which it is unclear how many results will be set.

    i.e. there may be speculation about when they can be considered complete.

    When the freeze method is called, no more values will be accepted and
    the DelayedResultVariable will behave as a normal ResultVariable.

    When a DelayedResultVariable is definitely full, it is freeze itself.

    DelayedResultVariable are queued with the scheduler at the point at which they might be complete.
    The scheduler can decide when to freeze them. A DelayedResultVariable  can be complete when
      - it contains enough elements
      - there are no providers which still have to provide some values (tracked inexactly)
        (a queue variable can be dequeued by the scheduler when a provider is added)
    """

    __slots__ = ("queued", "queues", "listeners")

    def __init__(self, queue: "QueueScheduler", value: T = None) -> None:
        ResultVariable.__init__(self, value)
        self.queued = False
        self.queues = queue
        if self.can_get():
            self.queue()

    def freeze(self) -> None:
        if self.hasValue:
            return
        self.queued = True
        self.hasValue = True
        for waiter in self.waiters:
            waiter.ready(self)
        # prevent memory leaks
        self.waiters = None
        self.listeners = None
        self.queues = None

    def queue(self) -> None:
        if self.queued:
            return
        self.queued = True
        self.queues.add_possible(self)

    def unqueue(self) -> None:
        self.queued = False

    def get_waiting_providers(self) -> int:
        """How many values are definitely still waiting for"""
        raise NotImplementedError()

    def get_progress_potential(self) -> int:
        """How many are actually waiting for us """
        return len(self.waiters)


ListValue = Union["Instance", List["Instance"]]


class Promise(IPromise[ListValue]):

    __slots__ = ("provider", "owner")

    def __init__(self, owner: "ListVariable", provider: "Statement"):
        self.provider: "Optional[Statement]" = provider
        self.owner: "ListVariable" = owner

    def set_value(self, value: ListValue, location: Location, recur: bool = True) -> None:
        self.owner.set_promised_value(self, value, location, recur)


class BaseListVariable(DelayedResultVariable[ListValue]):
    """
    List variable, but only the part that is independent of an instance
    """

    value: "List[Instance]"

    __slots__ = ("promisses", "done_promisses")

    def __init__(self, queue: "QueueScheduler") -> None:
        self.promisses: List[Promise] = []
        self.done_promisses: List[Promise] = []
        self.listeners: List[ResultCollector[ListValue]] = []
        super().__init__(queue, [])

    def get_promise(self, provider: "Statement") -> IPromise[ListValue]:
        out = Promise(self, provider)
        self.promisses.append(out)
        return out

    def set_promised_value(self, promis: Promise, value: ListValue, location: Location, recur: bool = True) -> None:
        self.done_promisses.append(promis)
        self.set_value(value, location, recur)

    def get_waiting_providers(self) -> int:
        # todo: optimize?
        out = len(self.promisses) - len(self.done_promisses)
        if out < 0:
            raise Exception("SEVERE: COMPILER STATE CORRUPT: provide count negative")
        return out

    def _set_value(self, value: ListValue, location: Location, recur: bool = True) -> bool:
        """
        First half of set_value, returns True if second half should run
        """
        if self.hasValue:
            if isinstance(value, list):
                if len(value) == 0:
                    # empty list terminates list addition
                    return False
                for subvalue in value:
                    if subvalue not in self.value:
                        raise ModifiedAfterFreezeException(
                            self,
                            instance=self.myself,
                            attribute=self.attribute,
                            value=value,
                            location=location,
                            reverse=not recur,
                        )
            elif value in self.value:
                return False
            else:
                raise ModifiedAfterFreezeException(
                    self, instance=self.myself, attribute=self.attribute, value=value, location=location, reverse=not recur
                )

        if isinstance(value, list):
            if len(value) == 0:
                # the values of empty lists need no processing,
                # but a set_value from an empty list may fulfill a promise, allowing this object to be queued
                if self.can_get():
                    self.queue()
            else:
                for v in value:
                    self.set_value(v, location, recur)
            return False

        if self.type is not None:
            self.type.validate(value)

        if value in self.value:
            # any set_value may fulfill a promise, allowing this object to be queued
            if self.can_get():
                self.queue()
            return False

        self.value.append(value)

        for listener in self.listeners:
            listener.receive_result(value, location)

        return True

    def set_value(self, value: ListValue, location: Location, recur: bool = True) -> None:
        if not self._set_value(value, location, recur):
            return
        if self.can_get():
            self.queue()

    def can_get(self) -> bool:
        return self.get_waiting_providers() == 0

    def get_progress_potential(self) -> int:
        """How many are actually waiting for us """
        return len(self.waiters) - len(self.listeners)

    def receive_result(self, value: ListValue, location: Location) -> None:
        self.set_value(value, location)

    def listener(self, resultcollector: ResultCollector, location: Location) -> None:
        for value in self.value:
            resultcollector.receive_result(value, location)
        if not self.hasValue:
            self.listeners.append(resultcollector)

    def is_multi(self) -> bool:
        return True

    def __str__(self) -> str:
        return "BaseListVariable %s" % (self.value)


class TempListVariable(BaseListVariable):

    __slots__ = ()

    def set_promised_value(self, promis: Promise, value: ListValue, location: Location, recur: bool = True) -> None:
        super().set_promised_value(promis, value, location, recur)
        # 100% accurate promisse tracking
        if len(self.promisses) == len(self.done_promisses):
            self.freeze()


class ListVariable(BaseListVariable):

    value: "List[Instance]"

    __slots__ = ("attribute", "myself")

    def __init__(self, attribute: "RelationAttribute", instance: "Instance", queue: "QueueScheduler") -> None:
        self.attribute: "RelationAttribute" = attribute
        self.myself = instance
        super().__init__(queue)

    def set_value(self, value: ListValue, location: Location, recur: bool = True) -> None:
        try:
            if not self._set_value(value, location, recur):
                return
        except ModifiedAfterFreezeException as e:
            if len(self.value) == self.attribute.high:
                new_exception: CompilerException = RuntimeException(
                    None, "Exceeded relation arity on attribute '%s' of instance '%s'" % (self.attribute.name, self.myself)
                )
                new_exception.set_location(location)
                raise new_exception
            raise e
        # set counterpart
        if self.attribute.end is not None and recur:
            value.set_attribute(self.attribute.end.name, self.myself, location, False)

        if self.attribute.high is not None:
            if len(self.value) > self.attribute.high:
                raise RuntimeException(
                    None, "List over full: max nr of items is %d, content is %s" % (self.attribute.high, self.value)
                )

            if self.attribute.high == len(self.value):
                self.freeze()

        if self.can_get():
            self.queue()

    def can_get(self) -> bool:
        return len(self.value) >= self.attribute.low and self.get_waiting_providers() == 0

    def __str__(self) -> str:
        return "ListVariable %s %s = %s" % (self.myself, self.attribute, self.value)


class OptionVariable(DelayedResultVariable["Instance"]):

    __slots__ = ("attribute", "myself", "location")

    def __init__(self, attribute: "Attribute", instance: "Instance", queue: "QueueScheduler") -> None:
        DelayedResultVariable.__init__(self, queue)
        self.value = None
        self.attribute = attribute
        self.myself = instance
        self.location = None

    def set_value(self, value: "Instance", location: Location, recur: bool = True) -> None:
        assert location is not None
        if self.hasValue:
            if self.value is None:
                raise ModifiedAfterFreezeException(
                    self, instance=self.myself, attribute=self.attribute, value=value, location=location, reverse=not recur
                )
            if self.value != value:
                raise DoubleSetException(self, None, value, location)

        if not isinstance(value, Unknown) and self.type is not None:
            self.type.validate(value)

        # set counterpart
        if self.attribute.end is not None and recur:
            value.set_attribute(self.attribute.end.name, self.myself, location, False)

        self.value = value
        self.location = location
        self.freeze()

    def get_waiting_providers(self) -> int:
        # todo: optimize?
        if self.provider is None:
            return 0
        if self.hasValue:
            return 0
        return 1

    def can_get(self) -> bool:
        return self.get_waiting_providers() == 0

    def get_value(self) -> "Instance":
        result = DelayedResultVariable.get_value(self)
        if result is None:
            raise OptionalValueException(self.myself, self.attribute)
        return result

    def __str__(self) -> str:
        return "OptionVariable %s %s = %s" % (self.myself, self.attribute, self.value)


class DeprecatedOptionVariable(OptionVariable):
    """
    Represents nullable attributes. In the future this class can be removed, and a standard
    ResultVariable with nullable type should be used.
    """

    def freeze(self) -> None:
        if self.value is None:
            warning: CompilerDeprecationWarning = CompilerDeprecationWarning(
                None,
                "No value for attribute %s.%s. Assign null instead of leaving unassigned." % (self.myself.type, self.attribute),
            )
            warning.set_location(self.myself.get_location())
            inmanta_warnings.warn(warning)
        super().freeze()


class QueueScheduler(object):
    """
    Object representing the compiler to the AST nodes. It provides access to the queueing mechanism and the type system.

    MUTABLE!
    """

    __slots__ = ("compiler", "runqueue", "waitqueue", "types", "allwaiters")

    def __init__(
        self,
        compiler: "Compiler",
        runqueue: "Deque[Waiter]",
        waitqueue: Deque[DelayedResultVariable],
        types: Dict[str, Type],
        allwaiters: "Set[Waiter]",
    ) -> None:
        self.compiler = compiler
        self.runqueue = runqueue
        self.waitqueue = waitqueue
        self.types = types
        self.allwaiters = allwaiters

    def add_running(self, item: "Waiter") -> None:
        self.runqueue.append(item)

    def add_possible(self, rv: DelayedResultVariable) -> None:
        self.waitqueue.append(rv)

    def get_compiler(self) -> "Compiler":
        return self.compiler

    def get_types(self) -> Dict[str, Type]:
        return self.types

    def add_to_all(self, item: "Waiter") -> None:
        self.allwaiters.add(item)

    def remove_from_all(self, item: "Waiter") -> None:
        self.allwaiters.remove(item)

    def get_tracker(self) -> Optional[Tracker]:
        return None

    def for_tracker(self, tracer: Tracker) -> "QueueScheduler":
        return DelegateQueueScheduler(self, tracer)


class DelegateQueueScheduler(QueueScheduler):
    __slots__ = ("__delegate", "__tracker")

    def __init__(self, delegate: QueueScheduler, tracker: Tracker):
        self.__delegate = delegate
        self.__tracker = tracker

    def add_running(self, item: "Waiter") -> None:
        self.__delegate.add_running(item)

    def add_possible(self, rv: DelayedResultVariable) -> None:
        self.__delegate.add_possible(rv)

    def get_compiler(self) -> "Compiler":
        return self.__delegate.get_compiler()

    def get_types(self) -> Dict[str, Type]:
        return self.__delegate.get_types()

    def add_to_all(self, item: "Waiter") -> None:
        self.__delegate.add_to_all(item)

    def remove_from_all(self, item: "Waiter") -> None:
        self.__delegate.remove_from_all(item)

    def get_tracker(self) -> Tracker:
        return self.__tracker

    def for_tracker(self, tracer: Tracker) -> QueueScheduler:
        return DelegateQueueScheduler(self.__delegate, tracer)


class Waiter(object):
    """
    Waiters represent an executable unit, that can be executed the result variables they depend on have their values.
    """

    __slots__ = ("waitcount", "queue", "done", "requires")

    requires: Dict[object, ResultVariable]

    def __init__(self, queue: QueueScheduler):
        self.waitcount = 1
        self.queue = queue
        self.queue.add_to_all(self)
        self.done = False

    def requeue_with_additional_requires(self, key: Hashable, waitable: ResultVariable) -> None:
        """
        Re-queue with an additional requirement
        """
        self.requires[key] = waitable
        self.waitfor(waitable)

    def waitfor(self, waitable: ResultVariable) -> None:
        self.waitcount = self.waitcount + 1
        waitable.waitfor(self)

    def ready(self, other: Union[ResultVariable, "Waiter"]) -> None:
        self.waitcount = self.waitcount - 1
        if self.waitcount == 0:
            self.queue.add_running(self)
        if self.waitcount < 0:
            raise Exception("SEVERE: COMPILER STATE CORRUPT: waitcount negative")

    def execute(self) -> None:
        pass


class ExecutionUnit(Waiter):
    """
    Basic assign statement:
     - Wait for a dict of requirements
     - Call the execute method on the expression, with a map of the resulting values
     - Assign the resulting value to the result variable

     @param provides: Whether to register this XU as provider to the result variable
    """

    __slots__ = ("result", "expression", "resolver", "queue_scheduler", "owner")

    def __init__(
        self,
        queue_scheduler: QueueScheduler,
        resolver: "Resolver",
        result: ResultVariable,
        requires: Dict[object, ResultVariable],
        expression: "ExpressionStatement",
        owner: "Optional[Statement]" = None,
    ):
        Waiter.__init__(self, queue_scheduler)
        self.result = result.get_promise(expression)
        self.requires = requires
        self.expression = expression
        self.resolver = resolver
        self.queue_scheduler = queue_scheduler
        for r in requires.values():
            self.waitfor(r)
        self.ready(self)
        if owner is not None:
            self.owner = owner
        else:
            self.owner = expression

    def _unsafe_execute(self) -> None:
        requires = {k: v.get_value() for (k, v) in self.requires.items()}
        value = self.expression.execute(requires, self.resolver, self.queue_scheduler)
        self.result.set_value(value, self.expression.location)
        self.done = True

    def execute(self) -> None:
        try:
            self._unsafe_execute()
        except RuntimeException as e:
            e.set_statement(self.owner)
            raise e

    def __repr__(self) -> str:
        return repr(self.expression)


class HangUnit(Waiter):
    """
    Wait for a dict of requirements, call the resume method on the resumer, with a map of the resulting values
    """

    __slots__ = ("resolver", "resumer", "target")

    def __init__(
        self,
        queue_scheduler: QueueScheduler,
        resolver: "Resolver",
        requires: Dict[object, ResultVariable],
        target: Optional[ResultVariable],
        resumer: "Resumer",
    ) -> None:
        Waiter.__init__(self, queue_scheduler)
        self.resolver = resolver
        self.requires = requires
        self.resumer = resumer
        self.target = target
        for r in requires.values():
            self.waitfor(r)
        self.ready(self)

    def execute(self) -> None:
        try:
            self.resumer.resume({k: v.get_value() for (k, v) in self.requires.items()}, self.resolver, self.queue, self.target)
        except RuntimeException as e:
            e.set_statement(self.resumer)
            raise e
        self.done = True


class RawUnit(Waiter):
    """
    Wait for a map of requirements, call the resume method on the resumer,
    but with a map of ResultVariables instead of their values
    """

    __slots__ = ("resolver", "resumer")

    def __init__(
        self,
        queue_scheduler: QueueScheduler,
        resolver: "Resolver",
        requires: Dict[object, ResultVariable],
        resumer: "RawResumer",
    ) -> None:
        Waiter.__init__(self, queue_scheduler)
        self.resolver = resolver
        self.requires = requires
        self.resumer = resumer
        for r in requires.values():
            self.waitfor(r)
        self.ready(self)

    def execute(self) -> None:
        try:
            self.resumer.resume(self.requires, self.resolver, self.queue)
        except RuntimeException as e:
            e.set_statement(self.resumer)
            e.location = self.resumer.location
            raise e
        self.done = True


"""
Resolution

 - lexical scope (module it is in, then parent modules)
   - handled in NS and direct_lookup on XU
 - dynamic scope (entity, for loop,...)
   - resolvers
   - lex scope as arg or root

resolution
 - FQN:  to lexical scope (via parents) -> to its  NS (imports) -> to target NS -> directLookup
 - N: to resolver -> to parents -> to NS (lex root) -> directlookup

"""

Typeorvalue = Union[Type, ResultVariable]


class Resolver(object):

    __slots__ = ("namespace", "dataflow_graph")

    def __init__(self, namespace: Namespace, enable_dataflow_graph: bool = False) -> None:
        self.namespace = namespace
        self.dataflow_graph: Optional[DataflowGraph] = DataflowGraph(self) if enable_dataflow_graph else None

    def lookup(self, name: str, root: Namespace = None) -> Typeorvalue:
        # override lexial root
        # i.e. delegate to parent, until we get to the root, then either go to our root or lexical root of our caller
        if root is not None:
            ns = root
        else:
            ns = self.namespace

        return ns.lookup(name)

    def get_root_resolver(self) -> "Resolver":
        return self

    def for_namespace(self, namespace: Namespace) -> "Resolver":
        return NamespaceResolver(self, namespace)

    def get_dataflow_node(self, name: str) -> "dataflow.AssignableNodeReference":
        try:
            result_variable: Typeorvalue = self.lookup(name)
            assert isinstance(result_variable, ResultVariable)
            return result_variable.get_dataflow_node()
        except NotFoundException:
            # This block is only executed if the model contains a reference to an undefined variable.
            # Since we don't know in which scope it should be defined, we assume top scope.
            root_graph: Optional[DataflowGraph] = self.get_root_resolver().dataflow_graph
            assert root_graph is not None
            return root_graph.get_own_variable(name)


class NamespaceResolver(Resolver):

    __slots__ = ("parent", "root")

    def __init__(self, parent: Resolver, lecial_root: Namespace) -> None:
        self.parent = parent
        self.root = lecial_root
        self.dataflow_graph: Optional[DataflowGraph] = None
        if parent.dataflow_graph is not None:
            self.dataflow_graph = DataflowGraph(self, parent.dataflow_graph)

    def lookup(self, name: str, root: Namespace = None) -> Typeorvalue:
        if root is not None:
            return self.parent.lookup(name, root)
        return self.parent.lookup(name, self.root)

    def for_namespace(self, namespace: Namespace) -> "Resolver":
        return NamespaceResolver(self, namespace)

    def get_root_resolver(self) -> "Resolver":
        return self.parent.get_root_resolver()


class ExecutionContext(Resolver):

    __slots__ = ("block", "slots", "resolver")

    def __init__(self, block: "BasicBlock", resolver: Resolver):
        self.block = block
        self.slots: Dict[str, ResultVariable] = {n: ResultVariable() for n in block.get_variables()}
        self.resolver = resolver
        self.dataflow_graph: Optional[DataflowGraph] = None
        if resolver.dataflow_graph is not None:
            self.dataflow_graph = DataflowGraph(self, resolver.dataflow_graph)
            for name, var in self.slots.items():
                node_ref: dataflow.AssignableNodeReference = dataflow.AssignableNode(name).reference()
                var.set_dataflow_node(node_ref)

    def lookup(self, name: str, root: Namespace = None) -> Typeorvalue:
        if "::" in name:
            return self.resolver.lookup(name, root)
        if name in self.slots:
            return self.slots[name]
        return self.resolver.lookup(name, root)

    def direct_lookup(self, name: str) -> ResultVariable:
        if name in self.slots:
            return self.slots[name]
        else:
            raise NotFoundException(None, name, "variable %s not found" % name)

    def emit(self, queue: QueueScheduler) -> None:
        self.block.emit(self, queue)

    def get_root_resolver(self) -> Resolver:
        return self.resolver.get_root_resolver()

    def for_namespace(self, namespace: Namespace) -> Resolver:
        return NamespaceResolver(self, namespace)


# also extends locatable
class Instance(ExecutionContext):
    def set_location(self, location: Location) -> None:
        Locatable.set_location(self, location)
        self.locations.append(location)

    def get_location(self) -> Location:
        return Locatable.get_location(self)

    location = property(get_location, set_location)

    __slots__ = (
        "_location",
        "resolver",
        "type",
        "sid",
        "implementations",
        "trackers",
        "locations",
        "instance_node",
    )

    def __init__(
        self,
        mytype: "Entity",
        resolver: Resolver,
        queue: QueueScheduler,
        node: Optional["dataflow.InstanceNodeReference"] = None,
    ) -> None:
        Locatable.__init__(self)
        # ExecutionContext, Resolver -> this class only uses it as an "interface", so no constructor call!
        self.resolver = resolver.get_root_resolver()
        self.type = mytype
        self.slots: Dict[str, ResultVariable] = {
            n: mytype.get_attribute(n).get_new_result_variable(self, queue)
            # prune duplicates first because get_new_result_variable() has side effects
            for n in set(mytype.get_all_attribute_names())
        }

        # TODO: this is somewhat ugly. Is there a cleaner way to enforce this constraint
        assert (resolver.dataflow_graph is None) == (node is None)
        self.dataflow_graph: Optional[DataflowGraph] = None
        self.instance_node: Optional[dataflow.InstanceNodeReference] = node
        if self.instance_node is not None:
            self.dataflow_graph = DataflowGraph(self, resolver.dataflow_graph)
            for name, var in self.slots.items():
                var.set_dataflow_node(dataflow.InstanceAttributeNodeReference(self.instance_node.top_node(), name))

        self.slots["self"] = ResultVariable()
        self.slots["self"].set_value(self, None)
        if self.instance_node is not None:
            self_var_node: dataflow.AssignableNodeReference = dataflow.AssignableNode("__self__").reference()
            self_var_node.assign(self.instance_node, self, cast(DataflowGraph, self.dataflow_graph))
            self.slots["self"].set_dataflow_node(self_var_node)

        self.sid = id(self)
        self.implementations: "Set[Implementation]" = set()

        # see inmanta.ast.execute.scheduler.QueueScheduler
        self.trackers: List[Tracker] = []

        self.locations: List[Location] = []

    def get_type(self) -> "Entity":
        return self.type

    def set_attribute(self, name: str, value: object, location: Location, recur: bool = True) -> None:
        if name not in self.slots:
            raise NotFoundException(None, name, "cannot set attribute with name %s on type %s" % (name, str(self.type)))
        try:
            self.slots[name].set_value(value, location, recur)
        except RuntimeException as e:
            raise AttributeException(self, self, name, cause=e)

    def get_attribute(self, name: str) -> ResultVariable:
        try:
            return self.slots[name]
        except KeyError:
            raise NotFoundException(None, name, "could not find attribute with name: %s in type %s" % (name, self.type))

    def __repr__(self) -> str:
        return "%s %02x" % (self.type, self.sid)

    def __str__(self) -> str:
        return "%s (instantiated at %s)" % (self.type, ",".join([str(location) for location in self.get_locations()]))

    def add_implementation(self, impl: "Implementation") -> bool:
        if impl in self.implementations:
            return False
        self.implementations.add(impl)
        return True

    def final(self, excns: List[CompilerException]) -> None:
        """
        The object should be complete, freeze all attributes
        """
        if len(self.implementations) == 0:
            excns.append(RuntimeException(self, "Unable to select implementation for entity %s" % self.type.name))

        for k, v in self.slots.items():
            if not v.is_ready():
                if v.can_get():
                    v.freeze()
                else:
                    attr = self.type.get_attribute(k)
                    if attr.is_multi():
                        low = attr.low
                        # none for list attributes
                        # list for n-ary relations
                        length = 0 if v.value is None else len(v.value)
                        excns.append(
                            proxy.UnsetException(
                                "The object %s is not complete: attribute %s (%s) requires %d values but only %d are set"
                                % (self, k, attr.location, low, length),
                                self,
                                attr,
                            )
                        )
                    else:
                        excns.append(
                            proxy.UnsetException(
                                "The object %s is not complete: attribute %s (%s) is not set" % (self, k, attr.location),
                                self,
                                attr,
                            )
                        )

    def dump(self) -> None:
        print("------------ ")
        print(str(self))
        print("------------ ")
        for (n, v) in self.slots.items():
            if v.can_get():

                value = v.value
                print("%s\t\t%s" % (n, value))
            else:
                print("BAD: %s\t\t%s" % (n, v.provider))

    def verify_done(self) -> bool:
        for v in self.slots.values():
            if not v.can_get():
                return False
        return True

    def get_locations(self) -> List[Location]:
        return self.locations



"""
    Copyright 2019 Inmanta

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Contact: code@inmanta.com
"""
import asyncio
import logging
import uuid

import pytest

from inmanta import const, data, resources
from inmanta.server import SLICE_AGENT_MANAGER
from inmanta.util import get_compiler_version
from utils import LogSequence, _wait_until_deployment_finishes, no_error_in_logs, retry_limited, wait_until_logs_are_available


@pytest.mark.asyncio
async def test_get_facts(resource_container, client, clienthelper, environment, agent, caplog):
    """
    Test retrieving facts from the agent
    """
    env_id = environment

    resource_container.Provider.set("agent1", "key", "value")

    version = await clienthelper.get_version()

    resource_id_wov = "test::Resource[agent1,key=key]"
    resource_id = "%s,v=%d" % (resource_id_wov, version)

    resources = [{"key": "key", "value": "value", "id": resource_id, "requires": [], "purged": False, "send_event": False}]

    await clienthelper.put_version_simple(resources, version)

    result = await client.release_version(env_id, version, True, const.AgentTriggerMethod.push_full_deploy)
    assert result.code == 200

    result = await client.get_param(env_id, "length", resource_id_wov)
    assert result.code == 503

    env_uuid = uuid.UUID(env_id)
    params = await data.Parameter.get_list(environment=env_uuid, resource_id=resource_id_wov)
    while len(params) < 3:
        params = await data.Parameter.get_list(environment=env_uuid, resource_id=resource_id_wov)
        await asyncio.sleep(0.1)

    result = await client.get_param(env_id, "key1", resource_id_wov)
    assert result.code == 200
    no_error_in_logs(caplog)


@pytest.mark.asyncio
async def test_purged_facts(resource_container, client, clienthelper, agent, environment, no_agent_backoff, caplog):
    """
    Test if facts are purged when the resource is purged.
    """
    resource_container.Provider.set("agent1", "key", "value")

    version = await clienthelper.get_version()
    resource_id_wov = "test::Resource[agent1,key=key]"
    resource_id = "%s,v=%d" % (resource_id_wov, version)

    resources = [{"key": "key", "value": "value", "id": resource_id, "requires": [], "purged": False, "send_event": False}]

    await clienthelper.put_version_simple(resources, version)

    result = await client.release_version(environment, version, True, const.AgentTriggerMethod.push_full_deploy)
    assert result.code == 200

    result = await client.get_param(environment, "length", resource_id_wov)
    assert result.code == 503

    env_uuid = uuid.UUID(environment)
    params = await data.Parameter.get_list(environment=env_uuid, resource_id=resource_id_wov)
    while len(params) < 3:
        params = await data.Parameter.get_list(environment=env_uuid, resource_id=resource_id_wov)
        await asyncio.sleep(0.1)

    result = await client.get_param(environment, "key1", resource_id_wov)
    assert result.code == 200

    # Purge the resource
    version = await clienthelper.get_version()
    resources[0]["id"] = "%s,v=%d" % (resource_id_wov, version)
    resources[0]["purged"] = True
    result = await client.put_version(
        tid=environment,
        version=version,
        resources=resources,
        unknowns=[],
        version_info={},
        compiler_version=get_compiler_version(),
    )
    assert result.code == 200
    result = await client.release_version(environment, version, True, const.AgentTriggerMethod.push_full_deploy)
    assert result.code == 200

    result = await client.get_version(environment, version)
    assert result.code == 200

    await _wait_until_deployment_finishes(client, environment, version)

    result = await client.get_version(environment, version)
    assert result.result["model"]["done"] == len(resources)

    # The resource facts should be purged
    result = await client.get_param(environment, "length", resource_id_wov)
    assert result.code == 503

    no_error_in_logs(caplog)


@pytest.mark.asyncio
async def test_get_facts_extended(server, client, agent, clienthelper, resource_container, environment, caplog):
    """
    dryrun and deploy a configuration model automatically
    """
    agentmanager = server.get_slice(SLICE_AGENT_MANAGER)
    # allow very rapid fact refresh
    agentmanager._fact_resource_block = 0.1

    resource_container.Provider.reset()

    version = await clienthelper.get_version()

    # mark some as existing
    resource_container.Provider.set("agent1", "key1", "value")
    resource_container.Provider.set("agent1", "key2", "value")
    resource_container.Provider.set("agent1", "key4", "value")
    resource_container.Provider.set("agent1", "key5", "value")
    resource_container.Provider.set("agent1", "key6", "value")
    resource_container.Provider.set("agent1", "key7", "value")

    resources = [
        {
            "key": "key1",
            "value": "value1",
            "id": "test::Fact[agent1,key=key1],v=%d" % version,
            "send_event": False,
            "purged": False,
            "skip": True,
            "skipFact": False,
            "factvalue": "fk1",
            "requires": [],
        },
        {
            "key": "key2",
            "value": "value1",
            "id": "test::Fact[agent1,key=key2],v=%d" % version,
            "send_event": False,
            "purged": False,
            "skip": False,
            "skipFact": False,
            "factvalue": "fk2",
            "requires": [],
        },
        {
            "key": "key3",
            "value": "value1",
            "id": "test::Fact[agent1,key=key3],v=%d" % version,
            "send_event": False,
            "purged": False,
            "skip": False,
            "skipFact": False,
            "factvalue": "fk3",
            "requires": [],
        },
        {
            "key": "key4",
            "value": "value1",
            "id": "test::Fact[agent1,key=key4],v=%d" % version,
            "send_event": False,
            "purged": False,
            "skip": False,
            "skipFact": False,
            "factvalue": "fk4",
            "requires": [],
        },
        {
            "key": "key5",
            "value": "value1",
            "id": "test::Fact[agent1,key=key5],v=%d" % version,
            "send_event": False,
            "purged": False,
            "skip": False,
            "skipFact": True,
            "factvalue": None,
            "requires": [],
        },
        {
            "key": "key6",
            "value": "value1",
            "id": "test::Fact[agent1,key=key6],v=%d" % version,
            "send_event": False,
            "purged": False,
            "skip": False,
            "skipFact": False,
            "factvalue": None,
            "requires": [],
        },
        {
            "key": "key7",
            "value": "value1",
            "id": "test::Fact[agent1,key=key7],v=%d" % version,
            "send_event": False,
            "purged": False,
            "skip": False,
            "skipFact": False,
            "factvalue": "",
            "requires": [],
        },
    ]

    resource_states = {
        "test::Fact[agent1,key=key4],v=%d" % version: const.ResourceState.undefined,
        "test::Fact[agent1,key=key5],v=%d" % version: const.ResourceState.undefined,
    }

    async def get_fact(rid, result_code=200, limit=10, lower_limit=2):
        lower_limit = limit - lower_limit
        result = await client.get_param(environment, "fact", rid)

        # add minimal nr of reps or failure cases
        while (result.code != result_code and limit > 0) or limit > lower_limit:
            limit -= 1
            await asyncio.sleep(0.1)
            result = await client.get_param(environment, "fact", rid)

        assert result.code == result_code
        return result

    result = await client.put_version(
        tid=environment,
        version=version,
        resources=resources,
        unknowns=[],
        version_info={},
        resource_state=resource_states,
        compiler_version=get_compiler_version(),
    )
    assert result.code == 200

    await get_fact("test::Fact[agent1,key=key1]")  # undeployable
    await get_fact("test::Fact[agent1,key=key2]")  # normal
    await get_fact("test::Fact[agent1,key=key3]", 503)  # not present
    await get_fact("test::Fact[agent1,key=key4]")  # unknown
    await get_fact("test::Fact[agent1,key=key5]", 503)  # broken
    f6 = await get_fact("test::Fact[agent1,key=key6]")  # normal
    f7 = await get_fact("test::Fact[agent1,key=key7]")  # normal

    assert f6.result["parameter"]["value"] == "None"
    assert f7.result["parameter"]["value"] == ""

    result = await client.release_version(environment, version, True, const.AgentTriggerMethod.push_full_deploy)
    assert result.code == 200

    await _wait_until_deployment_finishes(client, environment, version)

    await get_fact("test::Fact[agent1,key=key1]")  # undeployable
    await get_fact("test::Fact[agent1,key=key2]")  # normal
    await get_fact("test::Fact[agent1,key=key3]")  # not present -> present
    await get_fact("test::Fact[agent1,key=key4]")  # unknown
    await get_fact("test::Fact[agent1,key=key5]", 503)  # broken

    await agent.stop()

    LogSequence(caplog, allow_errors=False, ignore=["tornado.access"]).contains(
        "inmanta.agent.agent.agent1", logging.ERROR, "Unable to retrieve fact"
    ).contains("inmanta.agent.agent.agent1", logging.ERROR, "Unable to retrieve fact").contains(
        "inmanta.agent.agent.agent1", logging.ERROR, "Unable to retrieve fact"
    ).contains(
        "inmanta.agent.agent.agent1", logging.ERROR, "Unable to retrieve fact"
    ).contains(
        "inmanta.agent.agent.agent1", logging.ERROR, "Unable to retrieve fact"
    ).no_more_errors()


@pytest.mark.asyncio
async def test_purged_resources(resource_container, client, clienthelper, server, environment, agent, no_agent_backoff):
    """
    Test if:
        * Facts are purged when the resource is no longer available in any version.
        * Parameters with no resource attached are not deleted when a version is deleted.
    """
    resource_container.Provider.reset()

    resource_container.Provider.set("agent1", "key1", "value")

    res1 = "test::Resource[agent1,key=key1]"
    res2 = "test::Resource[agent1,key=key2]"

    # Add a resource independent parameter
    result = await client.set_param(
        tid=environment, id="test", source=const.ParameterSource.user, value="val", resource_id=None
    )
    assert result.code == 200

    # Create version 1 with 1 unknown
    version = await clienthelper.get_version()
    assert version == 1

    resources = [
        {"key": "key1", "value": "value", "id": f"{res1},v={version}", "requires": [], "purged": False, "send_event": False},
        {"key": "key2", "value": "value", "id": f"{res2},v={version}", "requires": [], "purged": False, "send_event": False},
    ]

    await clienthelper.put_version_simple(resources, version)

    result = await client.release_version(environment, version, True, const.AgentTriggerMethod.push_full_deploy)
    assert result.code == 200

    await _wait_until_deployment_finishes(client, environment, version)

    # Make sure we get facts
    result = await client.get_param(environment, "length", res1)
    assert result.code == 503

    result = await client.get_param(environment, "length", res2)
    assert result.code == 503

    env_uuid = uuid.UUID(environment)

    async def params_are_available() -> bool:
        params = await data.Parameter.get_list(environment=env_uuid)
        # 3 facts from res1 + 3 facts from res2 + parameter test
        return len(params) >= 7

    await retry_limited(params_are_available, 10)

    result = await client.get_param(environment, "key1", res1)
    assert result.code == 200

    result = await client.get_param(environment, "key2", res2)
    assert result.code == 200

    result = await client.get_param(environment, "test")
    assert result.code == 200

    # Create version 2
    version = await clienthelper.get_version()
    assert version == 2

    resources = [
        {"key": "key1", "value": "value", "id": f"{res1},v={version}", "requires": [], "purged": False, "send_event": False}
    ]

    await clienthelper.put_version_simple(resources, version)

    result = await client.release_version(environment, version, True, const.AgentTriggerMethod.push_full_deploy)
    assert result.code == 200

    # Remove version 1
    result = await client.delete_version(tid=environment, id=1)
    assert result.code == 200

    # There should be only one resource
    result = await client.get_environment(id=environment, resources=1)
    assert result.code == 200
    assert len(result.result["environment"]["resources"]) == 1

    # 3 facts from res1 + parameter test
    result = await client.list_params(environment)
    assert result.code == 200
    assert len(result.result["parameters"]) == 4

    # Remove version 2
    result = await client.delete_version(tid=environment, id=2)
    assert result.code == 200

    # Verify there are no resources anymore
    result = await client.get_environment(id=environment, resources=1)
    assert result.code == 200
    assert len(result.result["environment"]["resources"]) == 0

    # Only the resource independent parameter test should exist
    result = await client.list_params(environment)
    assert result.code == 200
    assert len(result.result["parameters"]) == 1


@pytest.mark.asyncio
async def test_get_fact_no_code(resource_container, client, clienthelper, environment, agent):
    """
    Test retrieving facts from the agent when resource cannot be loaded
    """
    env_id = environment
    resource_container.Provider.set("agent1", "key", "value")

    version = await clienthelper.get_version()

    resource_id_wov = "test::Resource[agent1,key=key]"
    resource_id = "%s,v=%d" % (resource_id_wov, version)

    await clienthelper.put_version_simple(
        [{"key": "key", "value": "value", "id": resource_id, "requires": [], "purged": False, "send_event": False}], version
    )

    response = await client.release_version(env_id, version, True, const.AgentTriggerMethod.push_full_deploy)
    assert response.code == 200

    await _wait_until_deployment_finishes(client, environment, version)

    # make sure the code to load the resource is no longer available
    resources.resource.reset()

    response = await client.get_param(env_id, "length", resource_id_wov)
    assert response.code == 503

    # The resource state and its logs are not set atomically. This call prevents a race condition.
    await wait_until_logs_are_available(client, environment, resource_id, expect_nr_of_logs=4)

    response = await client.get_resource(environment, resource_id, logs=True)
    assert response.code == 200
    result = response.result
    assert result["resource"]["status"] == "deployed"
    log_entry = result["logs"][0]
    assert log_entry["action"] == "getfact"
    assert log_entry["status"] == "unavailable"
    assert "Failed to load" in log_entry["messages"][0]["msg"]



"""
    Copyright 2019 Inmanta

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Contact: code@inmanta.com
"""
# Adapted from pyformance

import asyncio
import base64
import logging
import time
from asyncio import Task
from typing import Dict, Optional
from urllib.parse import quote

from pyformance import MetricsRegistry, global_registry
from tornado.httpclient import AsyncHTTPClient, HTTPError, HTTPRequest

LOGGER = logging.getLogger(__name__)

DEFAULT_INFLUX_SERVER = "127.0.0.1"
DEFAULT_INFLUX_PORT = 8086
DEFAULT_INFLUX_DATABASE = "metrics"
DEFAULT_INFLUX_USERNAME = None
DEFAULT_INFLUX_PASSWORD = None
DEFAULT_INFLUX_PROTOCOL = "http"


class AsyncReporter(object):
    def __init__(self, registry: MetricsRegistry = None, reporting_interval: int = 30) -> None:
        self.registry = registry or global_registry()
        self.reporting_interval = reporting_interval
        self._stopped = False
        self._handle: Optional[Task[None]] = None

    def start(self) -> bool:
        if self._stopped:
            return False
        self._handle = asyncio.get_event_loop().create_task(self._loop())
        return True

    def stop(self) -> None:
        self._stopped = True
        if self._handle is not None:
            self._handle.cancel()

    async def _loop(self) -> None:
        loop = asyncio.get_event_loop()
        next_loop_time = loop.time()
        while not self._stopped:
            try:
                await self.report_now(self.registry)
            except Exception:
                LOGGER.warning("Could not send metrics report", exc_info=True)
            next_loop_time += self.reporting_interval
            wait = max(0, next_loop_time - loop.time())
            await asyncio.sleep(wait)

    async def report_now(self, registry: MetricsRegistry = None, timestamp: float = None) -> None:
        raise NotImplementedError(self.report_now)


class InfluxReporter(AsyncReporter):

    """
    InfluxDB reporter using native http api
    (based on https://influxdb.com/docs/v1.1/guides/writing_data.html)
    """

    def __init__(
        self,
        registry: Optional[MetricsRegistry] = None,
        reporting_interval: int = 5,
        database: str = DEFAULT_INFLUX_DATABASE,
        server: str = DEFAULT_INFLUX_SERVER,
        username: Optional[str] = DEFAULT_INFLUX_USERNAME,
        password: Optional[str] = DEFAULT_INFLUX_PASSWORD,
        port: int = DEFAULT_INFLUX_PORT,
        protocol: str = DEFAULT_INFLUX_PROTOCOL,
        autocreate_database: bool = False,
        tags: Dict[str, str] = {},
    ) -> None:
        super(InfluxReporter, self).__init__(registry, reporting_interval)
        self.database = database
        self.username = username
        self.password = password
        self.port = port
        self.protocol = protocol
        self.server = server
        self.autocreate_database = autocreate_database
        self._did_create_database = False
        self.tags = tags
        self.key = "metrics"
        if self.tags:
            tagstring = ",".join("%s=%s" % (key, value) for key, value in self.tags.items())
            self.key = "%s,%s" % (self.key, tagstring)
        self.key = "%s,key=" % self.key

        if not self.server:
            raise Exception("Unable to start the metrics reporter without a server. Empty string given.")

    async def _create_database(self, http_client: AsyncHTTPClient) -> None:
        url = "%s://%s:%s/query" % (self.protocol, self.server, self.port)
        q = quote("CREATE DATABASE %s" % self.database)
        request = HTTPRequest(url + "?q=" + q)
        if self.username and self.password:
            auth = _encode_username(self.username, self.password)
            request.headers.add("Authorization", "Basic %s" % auth.decode("utf-8"))
        try:
            response = await http_client.fetch(request)
            response.rethrow()
            # Only set if we actually were able to get a successful response
            self._did_create_database = True
        except Exception:
            LOGGER.warning("Cannot create database %s to %s", self.database, self.server, exc_info=True)

    async def report_now(self, registry: Optional[MetricsRegistry] = None, timestamp: Optional[float] = None):
        http_client = AsyncHTTPClient()

        if self.autocreate_database and not self._did_create_database:
            await self._create_database(http_client)
        timestamp = timestamp or int(round(time.time()))
        metrics = (registry or self.registry).dump_metrics()
        post_data = []
        for key, metric_values in metrics.items():
            table = self.key + key
            values = ",".join(
                ["%s=%s" % (k, v if type(v) is not str else '"{}"'.format(v)) for (k, v) in metric_values.items()]
            )
            line = "%s %s %s" % (table, values, timestamp)
            post_data.append(line)
        post_data_all = "\n".join(post_data)
        path = "/write?db=%s&precision=s" % self.database
        url = "%s://%s:%s%s" % (self.protocol, self.server, self.port, path)
        request = HTTPRequest(url, method="POST", body=post_data_all.encode("utf-8"))
        if self.username and self.password:
            auth = _encode_username(self.username, self.password)
            request.headers.add("Authorization", "Basic %s" % auth.decode("utf-8"))
        try:
            response = await http_client.fetch(request)
            response.rethrow()
        except HTTPError:
            LOGGER.warning("Cannot write to %s", self.server, exc_info=True)


def _encode_username(username: str, password: str) -> bytes:
    auth_string = ("%s:%s" % (username, password)).encode()
    return base64.b64encode(auth_string)



"""
    Copyright 2019 Inmanta

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Contact: code@inmanta.com
"""
from collections import defaultdict

import pytest

from agent_server.conftest import get_agent, stop_agent
from inmanta import const, resources
from inmanta.agent import handler
from inmanta.agent.handler import CRUDHandler, HandlerContext, provider
from inmanta.export import unknown_parameters
from inmanta.loader import SourceInfo
from inmanta.resources import PurgeableResource, resource
from inmanta.types import JsonType
from utils import _wait_until_deployment_finishes


def reset_all_objects():
    resources.resource.reset()
    handler.Commander.reset()
    unknown_parameters.clear()


class Provider(CRUDHandler):
    def read_resource(self, ctx, current):
        self.read(current.id.get_agent_name(), current.key)
        assert current.value != const.UNKNOWN_STRING
        current.purged = not self.isset(current.id.get_agent_name(), current.key)

        if not current.purged:
            current.value = self.get(current.id.get_agent_name(), current.key)
        else:
            current.value = None

    def calculate_diff(self, ctx: HandlerContext, current: resources.Resource, desired: resources.Resource):
        diff = super().calculate_diff(ctx, current, desired)
        ctx.info("Diff was called")
        return diff

    def create_resource(self, ctx: HandlerContext, resource: PurgeableResource) -> None:
        self.touch(resource.id.get_agent_name(), resource.key)
        self.set(resource.id.get_agent_name(), resource.key, resource.value)
        ctx.set_created()

    def delete_resource(self, ctx: HandlerContext, resource: PurgeableResource) -> None:
        self.delete(resource.id.get_agent_name(), resource.key)
        ctx.set_purged()

    def update_resource(self, ctx: HandlerContext, changes: dict, resource: PurgeableResource) -> None:
        self.touch(resource.id.get_agent_name(), resource.key)
        self.set(resource.id.get_agent_name(), resource.key, resource.value)
        ctx.set_updated()

    _STATE = defaultdict(dict)
    _WRITE_COUNT = defaultdict(lambda: defaultdict(lambda: 0))
    _RELOAD_COUNT = defaultdict(lambda: defaultdict(lambda: 0))
    _READ_COUNT = defaultdict(lambda: defaultdict(lambda: 0))

    @classmethod
    def touch(cls, agent, key):
        cls._WRITE_COUNT[agent][key] += 1

    @classmethod
    def read(cls, agent, key):
        cls._READ_COUNT[agent][key] += 1

    @classmethod
    def set(cls, agent, key, value):
        cls._STATE[agent][key] = value

    @classmethod
    def get(cls, agent, key):
        if key in cls._STATE[agent]:
            return cls._STATE[agent][key]
        return None

    @classmethod
    def isset(cls, agent, key):
        return key in cls._STATE[agent]

    @classmethod
    def delete(cls, agent, key):
        if cls.isset(agent, key):
            del cls._STATE[agent][key]

    @classmethod
    def changecount(cls, agent, key):
        return cls._WRITE_COUNT[agent][key]

    @classmethod
    def readcount(cls, agent, key):
        return cls._READ_COUNT[agent][key]


def resource_container_a():
    reset_all_objects()

    @resource("__config__::Resource", agent="agent", id_attribute="key")
    class MyResource(PurgeableResource):
        """
        A file on a filesystem
        """

        fields = ("key", "value", "purged", "purge_on_delete")

    @provider("__config__::Resource", name="test_resource")
    class AProvider(Provider):
        pass

    return AProvider


def resource_container_b():
    reset_all_objects()

    @resource("__config__::Resource", agent="agent", id_attribute="uid")
    class MyResource(PurgeableResource):
        fields = ("uid", "key", "value", "purged", "purge_on_delete")

        def populate(self, fields: JsonType = None, force_fields: bool = False):
            if "uid" not in fields:
                # capture old format, cause by purge on delete
                fields["uid"] = None
            super().populate(fields, force_fields)

    @provider("__config__::Resource", name="test_resource")
    class BProvider(Provider):
        def read_resource(self, ctx, resource):
            if resource.uid is None:
                resource.purged = True
                return
            super().read_resource(ctx, resource)

        def create_resource(self, ctx: HandlerContext, resource: PurgeableResource) -> None:
            if resource.uid is None:
                return
            super().create_resource(ctx, resource)

        def delete_resource(self, ctx: HandlerContext, resource: PurgeableResource) -> None:
            if resource.uid is None:
                return
            super().delete_resource(ctx, resource)

        def update_resource(self, ctx: HandlerContext, changes: dict, resource: PurgeableResource) -> None:
            if resource.uid is None:
                return
            super().update_resource(ctx, resource)

    return BProvider


@pytest.mark.asyncio(timeout=150)
async def test_resource_evolution(server, client, environment, no_agent_backoff, snippetcompiler, monkeypatch, async_finalizer):

    provider = resource_container_a()

    agent = await get_agent(server, environment, "agent1")
    async_finalizer(agent.stop)

    # override origin check
    monkeypatch.setattr(SourceInfo, "_get_module_name", lambda s: s.module_name)
    monkeypatch.setattr(SourceInfo, "get_siblings", lambda s: iter([s]))
    monkeypatch.setattr(SourceInfo, "requires", [])

    snippetcompiler.setup_for_snippet(
        """
    entity Resource extends std::PurgeableResource:
        string key
        string value
        string agent
    end

    implement Resource using std::none

    Resource(key="a", value="b", agent="agent1")
    """
    )

    version, _ = await snippetcompiler.do_export_and_deploy()
    result = await client.release_version(environment, version, True, const.AgentTriggerMethod.push_full_deploy)
    assert result.code == 200

    await _wait_until_deployment_finishes(client, environment, version)

    await stop_agent(server, agent)

    assert provider.isset("agent1", "a")
    assert provider.changecount("agent1", "a") == 1

    provider = resource_container_b()

    agent = await get_agent(server, environment, "agent1")
    async_finalizer(agent.stop)

    snippetcompiler.setup_for_snippet(
        """
    entity Resource extends std::PurgeableResource:
        string key
        string value
        string agent
        string uid
    end

    implement Resource using std::none

    Resource(key="a", value="b", agent="agent1", uid="alpha")
    """
    )

    version, _ = await snippetcompiler.do_export_and_deploy()
    result = await client.release_version(environment, version, True, const.AgentTriggerMethod.push_full_deploy)
    assert result.code == 200
    assert result.result["model"]["total"] == 2

    await _wait_until_deployment_finishes(client, environment, version)
    assert provider.isset("agent1", "a")
    assert provider.changecount("agent1", "a") == 1

    snippetcompiler.reset()
    version, _ = await snippetcompiler.do_export_and_deploy()
    result = await client.release_version(environment, version, True, const.AgentTriggerMethod.push_full_deploy)
    assert result.code == 200
    assert result.result["model"]["total"] == 1

    await _wait_until_deployment_finishes(client, environment, version)
    assert provider.isset("agent1", "a")
    assert provider.changecount("agent1", "a") == 1



"""
    Copyright 2019 Inmanta

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Contact: code@inmanta.com
"""
import asyncio
import datetime
import logging
import os
import queue
import shutil
import subprocess
import uuid
from asyncio import Semaphore
from typing import AsyncIterator, List, Optional

import pytest

import inmanta.ast.export as ast_export
import inmanta.data.model as model
from inmanta import config, data
from inmanta.const import ParameterSource
from inmanta.data import APILIMIT, Compile, Report
from inmanta.deploy import cfg_env
from inmanta.protocol import Result
from inmanta.server import SLICE_COMPILER, SLICE_SERVER
from inmanta.server import config as server_config
from inmanta.server import protocol
from inmanta.server.bootloader import InmantaBootloader
from inmanta.server.protocol import Server
from inmanta.server.services.compilerservice import CompilerService, CompileRun, CompileStateListener
from inmanta.util import ensure_directory_exist
from utils import LogSequence, report_db_index_usage, retry_limited, wait_for_version

logger = logging.getLogger("inmanta.test.server.compilerservice")


@pytest.fixture
async def compilerservice(server_config, init_dataclasses_and_load_schema):
    server = Server()
    cs = CompilerService()
    await cs.prestart(server)
    await cs.start()
    yield cs
    await cs.prestop()
    await cs.stop()


@pytest.fixture
async def environment_factory(tmpdir) -> AsyncIterator["EnvironmentFactory"]:
    """
    Provides a factory for environments with a main.cf file.
    """
    yield EnvironmentFactory(str(tmpdir))


class EnvironmentFactory:
    def __init__(self, dir: str) -> None:
        self.src_dir: str = os.path.join(dir, "src")
        self.project: data.Project = data.Project(name="test")
        self._ready: bool = False

    async def setup(self) -> None:
        if self._ready:
            return

        project_template = os.path.join(os.path.dirname(os.path.abspath(__file__)), "..", "data", "project")
        shutil.copytree(project_template, self.src_dir)

        # Set up git
        subprocess.check_output(["git", "init"], cwd=self.src_dir)
        subprocess.check_output(["git", "add", "*"], cwd=self.src_dir)
        subprocess.check_output(["git", "config", "user.name", "Unit"], cwd=self.src_dir)
        subprocess.check_output(["git", "config", "user.email", "unit@test.example"], cwd=self.src_dir)
        subprocess.check_output(["git", "commit", "-m", "unit test"], cwd=self.src_dir)

        await self.project.insert()

        self._ready = True

    async def create_environment(self, main: str) -> data.Environment:
        await self.setup()
        branch: str = str(uuid.uuid4())
        subprocess.check_output(["git", "checkout", "-b", branch], cwd=self.src_dir)
        self.write_main(main)
        environment: data.Environment = data.Environment(
            name=branch, project=self.project.id, repo_url=self.src_dir, repo_branch=branch
        )
        await environment.insert()
        return environment

    def write_main(self, main: str, environment: Optional[data.Environment] = None) -> None:
        if environment is not None:
            subprocess.check_output(["git", "checkout", environment.repo_branch], cwd=self.src_dir)
        with open(os.path.join(self.src_dir, "main.cf"), "w", encoding="utf-8") as fd:
            fd.write(main)
        subprocess.check_output(["git", "add", "main.cf"], cwd=self.src_dir)
        subprocess.check_output(["git", "commit", "-m", "write main.cf", "--allow-empty"], cwd=self.src_dir)


@pytest.mark.asyncio
async def test_scheduler(server_config, init_dataclasses_and_load_schema, caplog):
    """Test the scheduler part in isolation, mock out compile runner and listen to state updates"""

    class Collector(CompileStateListener):
        """
        Collect all state updates, optionally hang the processing of listeners
        """

        def __init__(self):
            self.seen = []
            self.preseen = []
            self.lock = Semaphore(1)

        def reset(self):
            self.seen = []
            self.preseen = []

        async def compile_done(self, compile: data.Compile):
            self.preseen.append(compile)
            print("Got compile done for ", compile.remote_id)
            async with self.lock:
                self.seen.append(compile)

        async def hang(self):
            await self.lock.acquire()

        def release(self):
            self.lock.release()

        def verify(self, envs: uuid.UUID):
            assert sorted([x.remote_id for x in self.seen]) == sorted(envs)
            self.reset()

    class HangRunner(object):
        """
        compile runner mock, hang until released
        """

        def __init__(self):
            self.lock = Semaphore(0)
            self.started = False
            self.done = False
            self.version = None

        async def run(self, force_update: Optional[bool] = False):
            self.started = True
            await self.lock.acquire()
            self.done = True
            return True, None

        def release(self):
            self.lock.release()

    class HookedCompilerService(CompilerService):
        """
        hook in the hangrunner
        """

        def __init__(self):
            super(HookedCompilerService, self).__init__()
            self.locks = {}

        def _get_compile_runner(self, compile: data.Compile, project_dir: str):
            print("Get Run: ", compile.remote_id, compile.id)
            runner = HangRunner()
            self.locks[compile.remote_id] = runner
            return runner

        def get_runner(self, remote_id: uuid.UUID) -> HangRunner:
            return self.locks.get(remote_id)

    # manual setup of server
    server = Server()
    cs = HookedCompilerService()
    await cs.prestart(server)
    await cs.start()
    collector = Collector()
    cs.add_listener(collector)

    async def request_compile(env: data.Environment) -> uuid.UUID:
        """Request compile for given env, return remote_id"""
        u1 = uuid.uuid4()
        # add unique environment variables to prevent merging in request_recompile
        await cs.request_recompile(env, False, False, u1, env_vars={"uuid": str(u1)})
        results = await data.Compile.get_by_remote_id(env.id, u1)
        assert len(results) == 1
        assert results[0].remote_id == u1
        print("request: ", u1, results[0].id)
        return u1

    # setup projects in the database
    project = data.Project(name="test")
    await project.insert()
    env1 = data.Environment(name="dev", project=project.id, repo_url="", repo_branch="")
    await env1.insert()
    env2 = data.Environment(name="dev2", project=project.id, repo_url="", repo_branch="")
    await env2.insert()

    # setup series of compiles for two envs
    # e1 is for a plain run
    # e2 is for server restart
    e1 = [await request_compile(env1) for i in range(3)]
    e2 = [await request_compile(env2) for i in range(4)]
    print("env 1:", e1)

    async def check_compile_in_sequence(env: data.Environment, remote_ids: List[uuid.UUID], idx: int):
        """
        Check integrity of a compile sequence and progress the hangrunner.
        """
        before = remote_ids[:idx]

        for rid in before:
            prevrunner = cs.get_runner(rid)
            assert prevrunner.done

        if idx < len(remote_ids):
            current = remote_ids[idx]
            after = remote_ids[idx + 1 :]

            assert await cs.is_compiling(env.id) == 200

            await retry_limited(lambda: cs.get_runner(current) is not None, 1)
            await retry_limited(lambda: cs.get_runner(current).started, 1)

            for rid in after:
                nextrunner = cs.get_runner(rid)
                assert nextrunner is None

            cs.get_runner(current).release()
            await asyncio.sleep(0)
            await retry_limited(lambda: cs.get_runner(current).done, 1)

        else:

            async def isdone():
                return await cs.is_compiling(env.id) == 204

            await retry_limited(isdone, 1)

    # run through env1, entire sequence
    for i in range(4):
        await check_compile_in_sequence(env1, e1, i)
    collector.verify(e1)
    print("env1 done")

    print("env2 ", e2)
    # make event collector hang
    await collector.hang()
    # progress two steps into env2
    for i in range(2):
        await check_compile_in_sequence(env2, e2, i)

    assert not collector.seen
    print(collector.preseen)
    await retry_limited(lambda: len(collector.preseen) == 2, 1)

    # test server restart
    await cs.prestop()
    await cs.stop()

    # in the log, find cancel of compile(hangs) and handler(hangs)
    LogSequence(caplog, allow_errors=False).contains("inmanta.util", logging.WARNING, "was cancelled").contains(
        "inmanta.util", logging.WARNING, "was cancelled"
    ).no_more_errors()

    print("restarting")

    # restart new server
    cs = HookedCompilerService()
    await cs.prestart(server)
    await cs.start()
    collector = Collector()
    cs.add_listener(collector)

    # complete the sequence, expect re-run of third compile
    for i in range(3):
        print(i)
        await check_compile_in_sequence(env2, e2[2:], i)

    # all are re-run, entire sequence present
    collector.verify(e2)

    await report_db_index_usage()


@pytest.mark.asyncio
async def test_compile_runner(environment_factory: EnvironmentFactory, server, client, tmpdir):
    testmarker_env = "TESTMARKER"
    no_marker = "__no__marker__"
    marker_print = "_INM_MM:"
    marker_print2 = "_INM_MM2:"
    marker_print3 = "_INM_MM3:"

    def make_main(marker_print):
        return f"""
    marker = std::get_env("{testmarker_env}","{no_marker}")
    std::print("{marker_print} {{{{marker}}}}")

    host = std::Host(name="test", os=std::linux)
    std::ConfigFile(host=host, path="/etc/motd", content="1234")
        """

    env = await environment_factory.create_environment(make_main(marker_print))
    env2 = await environment_factory.create_environment(make_main(marker_print2))

    project_work_dir = os.path.join(tmpdir, "work")
    ensure_directory_exist(project_work_dir)

    async def compile_and_assert(env, export=True, meta={}, env_vars={}, update=False):
        compile = data.Compile(
            remote_id=uuid.uuid4(),
            environment=env.id,
            do_export=export,
            metadata=meta,
            environment_variables=env_vars,
            force_update=update,
        )
        await compile.insert()

        # compile with export
        cr = CompileRun(compile, project_work_dir)
        await cr.run()

        # get and process reports
        result = await client.get_report(compile.id)
        assert result.code == 200
        stages = result.result["report"]["reports"]
        if export:
            assert cr.version > 0
            result = await client.get_version(env.id, cr.version)
            assert result.code == 200
            metadata = result.result["model"]["version_info"]["export_metadata"]
            for k, v in meta.items():
                assert k in metadata
                assert v == metadata[k]

        print(stages)

        stage_by_name = {stage["name"]: stage for stage in stages}

        return cr, stage_by_name

    # with export
    compile, stages = await compile_and_assert(env, True, meta={"type": "Test"})
    assert stages["Init"]["returncode"] == 0
    assert stages["Cloning repository"]["returncode"] == 0
    assert stages["Recompiling configuration model"]["returncode"] == 0
    out = stages["Recompiling configuration model"]["outstream"]
    assert f"{marker_print} {no_marker}" in out
    assert len(stages) == 3
    assert compile.version is not None

    # no export
    compile, stages = await compile_and_assert(env, False)
    assert stages["Init"]["returncode"] == 0
    assert stages["Recompiling configuration model"]["returncode"] == 0
    out = stages["Recompiling configuration model"]["outstream"]
    assert f"{marker_print} {no_marker}" in out
    assert len(stages) == 2
    assert compile.version is None

    # env vars
    marker = str(uuid.uuid4())
    compile, stages = await compile_and_assert(env, False, env_vars={testmarker_env: marker})
    assert len(compile.request.environment_variables) == 1
    assert stages["Init"]["returncode"] == 0
    assert f"Using extra environment variables during compile TESTMARKER='{marker}'" in stages["Init"]["outstream"]
    assert stages["Recompiling configuration model"]["returncode"] == 0
    out = stages["Recompiling configuration model"]["outstream"]
    assert f"{marker_print} {marker}" in out
    assert len(stages) == 2
    assert compile.version is None

    # switch branch
    compile, stages = await compile_and_assert(env2, False)
    assert stages["Init"]["returncode"] == 0
    assert stages[f"switching branch from {env.repo_branch} to {env2.repo_branch}"]["returncode"] == 0
    assert stages["Recompiling configuration model"]["returncode"] == 0
    out = stages["Recompiling configuration model"]["outstream"]
    assert f"{marker_print2} {no_marker}" in out
    assert len(stages) == 3
    assert compile.version is None

    # update with no update
    compile, stages = await compile_and_assert(env2, False, update=True)
    assert stages["Init"]["returncode"] == 0
    assert stages["Fetching changes"]["returncode"] == 0
    assert stages["Pulling updates"]["returncode"] == 0
    assert stages["Updating modules"]["returncode"] == 0
    assert stages["Recompiling configuration model"]["returncode"] == 0
    out = stages["Recompiling configuration model"]["outstream"]
    assert f"{marker_print2} {no_marker}" in out
    assert len(stages) == 5
    assert compile.version is None

    environment_factory.write_main(make_main(marker_print3))
    compile, stages = await compile_and_assert(env2, False, update=True)
    assert stages["Init"]["returncode"] == 0
    assert stages["Fetching changes"]["returncode"] == 0
    assert stages["Pulling updates"]["returncode"] == 0
    assert stages["Updating modules"]["returncode"] == 0
    assert stages["Recompiling configuration model"]["returncode"] == 0
    out = stages["Recompiling configuration model"]["outstream"]
    assert f"{marker_print3} {no_marker}" in out
    assert len(stages) == 5
    assert compile.version is None


@pytest.mark.asyncio
async def test_compilerservice_compile_data(environment_factory: EnvironmentFactory, client, server) -> None:
    async def get_compile_data(main: str) -> model.CompileData:
        env: data.Environment = await environment_factory.create_environment(main)
        result: Result = await client.notify_change(env.id)
        assert result.code == 200

        async def compile_done():
            return (await client.is_compiling(env.id)).code == 204

        await retry_limited(compile_done, 10)

        reports = await client.get_reports(env.id)
        assert reports.code == 200
        assert len(reports.result["reports"]) == 1
        compile_id: str = reports.result["reports"][0]["id"]
        compile_data_a = model.CompileData(**reports.result["reports"][0]["compile_data"])

        result = await client.get_compile_data(uuid.UUID(compile_id))
        assert result.code == 200
        assert "data" in result.result
        compile_data: model.CompileData = model.CompileData(**result.result["data"])
        assert compile_data_a == compile_data
        return compile_data

    errors0: List[ast_export.Error] = (await get_compile_data("x = 0")).errors
    assert len(errors0) == 0

    errors1: List[ast_export.Error] = (await get_compile_data("x = 0 x = 1")).errors
    assert len(errors1) == 1
    error: ast_export.Error = errors1[0]
    assert error.category == ast_export.ErrorCategory.runtime
    assert error.type == "inmanta.ast.DoubleSetException"
    assert error.message == "value set twice:\n\told value: 0\n\t\tset at ./main.cf:1\n\tnew value: 1\n\t\tset at ./main.cf:1\n"


@pytest.mark.asyncio
async def test_e2e_recompile_failure(compilerservice: CompilerService):
    project = data.Project(name="test")
    await project.insert()

    env = data.Environment(name="dev", project=project.id, repo_url="", repo_branch="")
    await env.insert()

    u1 = uuid.uuid4()
    await compilerservice.request_recompile(env, False, False, u1, env_vars={"my_unique_var": str(u1)})
    u2 = uuid.uuid4()
    await compilerservice.request_recompile(env, False, False, u2, env_vars={"my_unique_var": str(u2)})

    assert await compilerservice.is_compiling(env.id) == 200

    async def compile_done():
        res = await compilerservice.is_compiling(env.id)
        print(res)
        return res == 204

    await retry_limited(compile_done, 10)

    _, all_compiles = await compilerservice.get_reports(env)
    all_reports = {i["remote_id"]: await compilerservice.get_report(i["id"]) for i in all_compiles["reports"]}

    def assert_report(uid):
        code, report = all_reports[uid]
        report = report["report"]
        assert report["remote_id"] == uid
        assert not report["success"]
        reports = report["reports"]
        reports = {r["name"]: r for r in reports}

        # stages
        init = reports["Init"]
        assert not init["errstream"]
        assert "project found in" in init["outstream"] and "and no repository set" in init["outstream"]

        # compile
        comp = reports["Recompiling configuration model"]
        assert "Unable to find an inmanta project (project.yml expected)" in comp["errstream"]
        assert comp["returncode"] == 1
        return report["requested"], report["started"], report["completed"]

    r1, s1, f1 = assert_report(u1)
    r2, s2, f2 = assert_report(u2)

    assert r2 > r1
    assert r1 < s1 < f1
    assert r2 < s2 < f2
    assert f1 < s2


@pytest.mark.asyncio(timeout=90)
async def test_server_recompile(server, client, environment, monkeypatch):
    """
    Test a recompile on the server and verify recompile triggers
    """
    config.Config.set("server", "auto-recompile-wait", "0")

    project_dir = os.path.join(server.get_slice(SLICE_SERVER)._server_storage["environments"], str(environment))
    project_source = os.path.join(os.path.dirname(os.path.abspath(__file__)), "..", "data", "project")
    print("Project at: ", project_dir)

    shutil.copytree(project_source, project_dir)
    subprocess.check_output(["git", "init"], cwd=project_dir)
    subprocess.check_output(["git", "add", "*"], cwd=project_dir)
    subprocess.check_output(["git", "config", "user.name", "Unit"], cwd=project_dir)
    subprocess.check_output(["git", "config", "user.email", "unit@test.example"], cwd=project_dir)
    subprocess.check_output(["git", "commit", "-m", "unit test"], cwd=project_dir)

    # Set environment variable to be passed to the compiler
    key_env_var = "TEST_MESSAGE"
    value_env_var = "a_message"
    monkeypatch.setenv(key_env_var, value_env_var)

    # add main.cf
    with open(os.path.join(project_dir, "main.cf"), "w", encoding="utf-8") as fd:
        fd.write(
            f"""
        host = std::Host(name="test", os=std::linux)
        std::ConfigFile(host=host, path="/etc/motd", content="1234")
        std::print(std::get_env("{key_env_var}"))
"""
        )

    logger.info("request a compile")
    result = await client.notify_change(environment)
    assert result.code == 200

    logger.info("wait for 1")
    versions = await wait_for_version(client, environment, 1)
    assert versions["versions"][0]["total"] == 1
    assert versions["versions"][0]["version_info"]["export_metadata"]["type"] == "api"

    # get compile reports and make sure the environment variables are not logged
    reports = await client.get_reports(environment)
    assert reports.code == 200
    assert len(reports.result["reports"]) == 1
    env_vars_compile = reports.result["reports"][0]["environment_variables"]
    assert key_env_var not in env_vars_compile

    # get report
    compile_report = await client.get_report(reports.result["reports"][0]["id"])
    assert compile_report.code == 200
    report_map = {r["name"]: r for r in compile_report.result["report"]["reports"]}
    assert value_env_var in report_map["Recompiling configuration model"]["outstream"]

    # set a parameter without requesting a recompile
    await client.set_param(environment, id="param1", value="test", source=ParameterSource.plugin)
    versions = await wait_for_version(client, environment, 1)
    assert versions["count"] == 1

    logger.info("request second compile")
    # set a new parameter and request a recompile
    await client.set_param(environment, id="param2", value="test", source=ParameterSource.plugin, recompile=True)
    logger.info("wait for 2")
    versions = await wait_for_version(client, environment, 2)
    assert versions["versions"][0]["version_info"]["export_metadata"]["type"] == "param"
    assert versions["count"] == 2

    # update the parameter to the same value -> no compile
    await client.set_param(environment, id="param2", value="test", source=ParameterSource.plugin, recompile=True)
    versions = await wait_for_version(client, environment, 2)
    assert versions["count"] == 2

    # update the parameter to a new value
    await client.set_param(environment, id="param2", value="test2", source=ParameterSource.plugin, recompile=True)
    versions = await wait_for_version(client, environment, 3)
    logger.info("wait for 3")
    assert versions["count"] == 3

    # clear the environment
    state_dir = server_config.state_dir.get()
    project_dir = os.path.join(state_dir, "server", "environments", environment)
    assert os.path.exists(project_dir)

    await client.clear_environment(environment)

    assert not os.path.exists(project_dir)


async def run_compile_and_wait_until_compile_is_done(
    compiler_service: CompilerService, compiler_queue: queue.Queue, env_id: uuid.UUID
) -> None:
    """
    Unblock the first compile in the compiler queue and wait until the compile finishes.
    """
    current_task = compiler_service._recompiles[env_id]

    # prevent race conditions where compile request is not yet in queue
    await retry_limited(lambda: not compiler_queue.empty(), timeout=10)
    run = compiler_queue.get(block=True)
    run.block = False

    def _is_compile_finished() -> bool:
        if env_id not in compiler_service._recompiles:
            return True
        if current_task is not compiler_service._recompiles[env_id]:
            return True
        return False

    await retry_limited(_is_compile_finished, timeout=10)


@pytest.mark.asyncio(timeout=90)
async def test_compileservice_queue(mocked_compiler_service_block: queue.Queue, server, client, environment):
    """
    Test the inspection of the compile queue. The compile runner is mocked out so the "started" field does not have the
    correct value in this test.
    """
    env = await data.Environment.get_by_id(environment)
    config.Config.set("server", "auto-recompile-wait", "0")
    compilerslice: CompilerService = server.get_slice(SLICE_COMPILER)

    result = await client.get_compile_queue(environment)
    assert len(result.result["queue"]) == 0
    assert result.code == 200

    # request a compile
    remote_id1 = uuid.uuid4()
    await compilerslice.request_recompile(
        env=env, force_update=False, do_export=False, remote_id=remote_id1, env_vars={"my_unique_var": "1"}
    )

    # api should return one
    result = await client.get_compile_queue(environment)
    assert len(result.result["queue"]) == 1
    assert result.result["queue"][0]["remote_id"] == str(remote_id1)
    assert result.code == 200

    # request a compile
    remote_id2 = uuid.uuid4()
    compile_id2, _ = await compilerslice.request_recompile(env=env, force_update=False, do_export=False, remote_id=remote_id2)

    # api should return two
    result = await client.get_compile_queue(environment)
    assert len(result.result["queue"]) == 2
    assert result.result["queue"][1]["remote_id"] == str(remote_id2)
    assert result.code == 200

    # request a compile with do_export=True
    remote_id3 = uuid.uuid4()
    compile_id3, _ = await compilerslice.request_recompile(env=env, force_update=False, do_export=True, remote_id=remote_id3)

    result = await client.get_compile_queue(environment)
    assert len(result.result["queue"]) == 3
    assert result.result["queue"][2]["remote_id"] == str(remote_id3)
    assert result.code == 200

    # request a compile with do_export=False -> expect merge with compile2
    remote_id4 = uuid.uuid4()
    compile_id4, _ = await compilerslice.request_recompile(env=env, force_update=False, do_export=False, remote_id=remote_id4)

    result = await client.get_compile_queue(environment)
    assert len(result.result["queue"]) == 4
    assert result.result["queue"][3]["remote_id"] == str(remote_id4)
    assert result.code == 200

    # request a compile with do_export=True -> expect merge with compile3, expect force_update == True for the compile
    remote_id5 = uuid.uuid4()
    compile_id5, _ = await compilerslice.request_recompile(env=env, force_update=True, do_export=True, remote_id=remote_id5)
    remote_id6 = uuid.uuid4()
    compile_id6, _ = await compilerslice.request_recompile(env=env, force_update=False, do_export=True, remote_id=remote_id6)

    result = await client.get_compile_queue(environment)
    assert len(result.result["queue"]) == 6
    assert result.result["queue"][4]["remote_id"] == str(remote_id5)
    assert result.result["queue"][5]["remote_id"] == str(remote_id6)
    assert result.code == 200

    # finish a compile and wait for service to take on next
    await run_compile_and_wait_until_compile_is_done(compilerslice, mocked_compiler_service_block, env.id)

    # api should return five when ready
    result = await client.get_compile_queue(environment)
    assert len(result.result["queue"]) == 5
    assert result.result["queue"][0]["remote_id"] == str(remote_id2)
    assert result.code == 200

    # finish second compile
    await run_compile_and_wait_until_compile_is_done(compilerslice, mocked_compiler_service_block, env.id)

    assert await compilerslice.get_report(compile_id2) == await compilerslice.get_report(compile_id4)

    # finish third compile
    # prevent race conditions where compile is not yet in queue
    await retry_limited(lambda: not mocked_compiler_service_block.empty(), timeout=10)
    run = mocked_compiler_service_block.get(block=True)
    result = await client.get_compile_queue(environment)
    assert len(result.result["queue"]) == 3
    assert result.result["queue"][0]["remote_id"] == str(remote_id3)
    assert result.code == 200
    run.block = False

    while env.id in compilerslice._recompiles:
        await asyncio.sleep(0.2)

    assert await compilerslice.get_report(compile_id3) == await compilerslice.get_report(compile_id5)
    assert await compilerslice.get_report(compile_id3) == await compilerslice.get_report(compile_id6)

    # api should return none
    result = await client.get_compile_queue(environment)
    assert len(result.result["queue"]) == 0
    assert result.code == 200


@pytest.mark.asyncio
async def test_compilerservice_halt(mocked_compiler_service_block, server, client, environment: uuid.UUID) -> None:
    config.Config.set("server", "auto-recompile-wait", "0")
    compilerslice: CompilerService = server.get_slice(SLICE_COMPILER)

    result = await client.get_compile_queue(environment)
    assert result.code == 200
    assert len(result.result["queue"]) == 0

    await client.halt_environment(environment)

    env = await data.Environment.get_by_id(environment)
    assert env is not None
    await compilerslice.request_recompile(env=env, force_update=False, do_export=False, remote_id=uuid.uuid4())

    result = await client.get_compile_queue(environment)
    assert result.code == 200
    assert len(result.result["queue"]) == 1

    result = await client.is_compiling(environment)
    assert result.code == 204

    await client.resume_environment(environment)
    result = await client.is_compiling(environment)
    assert result.code == 200


@pytest.fixture(scope="function")
async def server_with_frequent_cleanups(server_pre_start, server_config, async_finalizer):
    config.Config.set("server", "compiler-report-retention", "2")
    config.Config.set("server", "cleanup-compiler-reports_interval", "1")
    ibl = InmantaBootloader()
    await ibl.start()
    yield ibl.restserver
    await asyncio.wait_for(ibl.stop(), 15)


@pytest.fixture(scope="function")
def client_for_cleanup(server_with_frequent_cleanups):
    client = protocol.Client("client")
    yield client


@pytest.fixture(scope="function")
async def environment_for_cleanup(client_for_cleanup, server_with_frequent_cleanups):
    """
    Create a project and environment. This fixture returns the uuid of the environment
    """
    result = await client_for_cleanup.create_project("env-test")
    assert result.code == 200
    project_id = result.result["project"]["id"]

    result = await client_for_cleanup.create_environment(project_id=project_id, name="dev")
    env_id = result.result["environment"]["id"]

    cfg_env.set(env_id)

    yield env_id


@pytest.fixture
async def old_compile_report(server_with_frequent_cleanups, environment_for_cleanup):
    now = datetime.datetime.now()
    time_of_compile = now - datetime.timedelta(days=30)
    compile_id = uuid.UUID("c00cc33f-f70f-4800-ad01-ff042f67118f")
    old_compile = {
        "id": compile_id,
        "remote_id": uuid.UUID("c9a10da1-9bf6-4152-8461-98adc02c4cee"),
        "environment": uuid.UUID(environment_for_cleanup),
        "requested": time_of_compile,
        "started": time_of_compile,
        "completed": time_of_compile,
        "do_export": True,
        "force_update": True,
        "metadata": {"type": "api", "message": "Recompile trigger through API call"},
        "environment_variables": {},
        "success": True,
        "handled": True,
        "version": 1,
    }
    await Compile(**old_compile).insert()
    new_compile = {**old_compile, "id": uuid.uuid4(), "requested": now, "started": now, "completed": now}
    await Compile(**new_compile).insert()

    report1 = {
        "id": uuid.UUID("2baa0175-9169-40c5-a546-64d646f62da6"),
        "started": time_of_compile,
        "completed": time_of_compile,
        "command": "",
        "name": "Init",
        "errstream": "",
        "outstream": "Using extra environment variables during compile \n",
        "returncode": 0,
        "compile": compile_id,
    }
    report2 = {
        **report1,
        "id": uuid.UUID("2a86d2a0-666f-4cca-b0a8-13e0379128d5"),
        "command": "python -m inmanta.app -vvv export -X",
        "name": "Recompiling configuration model",
    }
    await Report(**report1).insert()
    await Report(**report2).insert()
    yield compile_id


@pytest.mark.asyncio
async def test_compileservice_cleanup(
    server_with_frequent_cleanups, client_for_cleanup, environment_for_cleanup, old_compile_report
):
    # There is a new and an older report in the database
    result = await client_for_cleanup.get_reports(environment_for_cleanup)
    assert result.code == 200
    assert len(result.result["reports"]) == 2

    result = await client_for_cleanup.get_report(old_compile_report)
    assert result.code == 200
    assert len(result.result["report"]["reports"]) > 0

    compilerslice: CompilerService = server_with_frequent_cleanups.get_slice(SLICE_COMPILER)
    await compilerslice._cleanup()

    # The old report is deleted after cleanup
    result = await client_for_cleanup.get_report(old_compile_report)
    assert result.code == 404
    reports_after_cleanup = await Report.get_list(compile=old_compile_report)
    assert len(reports_after_cleanup) == 0

    # The new report is still there
    result = await client_for_cleanup.get_reports(environment_for_cleanup)
    assert result.code == 200
    assert len(result.result["reports"]) == 1


@pytest.mark.slowtest
@pytest.mark.asyncio
async def test_compileservice_cleanup_on_trigger(client_for_cleanup, environment_for_cleanup, old_compile_report):
    # Two reports are in the table
    result = await client_for_cleanup.get_reports(environment_for_cleanup)
    assert result.code == 200
    assert len(result.result["reports"]) == 2

    result = await client_for_cleanup.get_report(old_compile_report)
    assert result.code == 200
    assert len(result.result["report"]["reports"]) > 0

    await asyncio.sleep(3)

    # Both reports should be deleted after the triggered cleanup
    result = await client_for_cleanup.get_reports(environment_for_cleanup)
    assert result.code == 200
    assert len(result.result["reports"]) == 0


@pytest.mark.asyncio
async def test_issue_2361(environment_factory: EnvironmentFactory, server, client, tmpdir):
    env = await environment_factory.create_environment(main="")

    # Change the branch of the environment to a non-existing branch as such that the run method
    # of the CompileRun returns after executing the clone stage.
    result = await client.environment_modify(id=env.id, name=env.id, branch="non-existing-branch")
    assert result.code == 200

    project_work_dir = os.path.join(tmpdir, "work")
    ensure_directory_exist(project_work_dir)

    compile = data.Compile(
        remote_id=uuid.uuid4(),
        environment=env.id,
        do_export=True,
        metadata={},
        environment_variables={},
        force_update=True,
    )
    await compile.insert()

    cr = CompileRun(compile, project_work_dir)

    # This should not result in a "local variable referenced before assignment" exception
    success, compile_data = await cr.run()
    assert not success
    assert compile_data is None


@pytest.mark.asyncio
async def test_git_uses_environment_variables(environment_factory: EnvironmentFactory, server, client, tmpdir, monkeypatch):
    """
    Make sure that the git clone command on the compilerservice takes into account the environment variables
    set on the system.
    """
    env = await environment_factory.create_environment(main="")
    result = await client.environment_modify(id=env.id, name=env.id, branch="master")
    assert result.code == 200

    project_work_dir = os.path.join(tmpdir, "work")
    ensure_directory_exist(project_work_dir)

    compile = data.Compile(
        remote_id=uuid.uuid4(),
        environment=env.id,
        do_export=True,
        metadata={},
        environment_variables={},
        force_update=True,
    )
    await compile.insert()

    # Make sure git clone logs trace lines
    monkeypatch.setenv("GIT_TRACE", "1")
    cr = CompileRun(compile, project_work_dir)
    await cr.run()

    report = await data.Report.get_one(compile=compile.id, name="Cloning repository")
    # Assert presence of trace lines
    assert "trace: " in report.errstream


@pytest.mark.asyncio
async def test_compileservice_api(client, environment):
    # Exceed max value for limit
    result = await client.get_reports(environment, limit=APILIMIT + 1)
    assert result.code == 400
    assert result.result["message"] == f"Invalid request: limit parameter can not exceed {APILIMIT}, got {APILIMIT+1}."

    result = await client.get_reports(environment, limit=APILIMIT)
    assert result.code == 200



"""
    Copyright 2016 Inmanta

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Contact: code@inmanta.com
"""
import asyncio
import datetime
import logging
import typing
from asyncio import subprocess
from typing import Dict, List, Optional, Set, Tuple
from unittest.mock import Mock
from uuid import UUID, uuid4

import pytest

from inmanta import config, data
from inmanta.agent import Agent, agent
from inmanta.agent import config as agent_config
from inmanta.const import AgentAction, AgentStatus
from inmanta.protocol import Result
from inmanta.server import SLICE_AGENT_MANAGER, SLICE_AUTOSTARTED_AGENT_MANAGER
from inmanta.server.agentmanager import AgentManager, AutostartedAgentManager, SessionAction, SessionManager
from inmanta.server.protocol import Session
from utils import UNKWN, assert_equal_ish, retry_limited


class Collector(object):
    def __init__(self):
        self.values = []

    def __call__(self, arg):
        self.values.append(arg)

    async def proccess(self):
        while len(self.values) > 0:
            x = self.values
            self.values = []
            await asyncio.gather(*x)


async def empty_future(*args, **kwargs):
    pass


class MockSession(object):
    """
    An environment that segments agents connected to the server
    """

    def __init__(self, sid, tid, endpoint_names: Set[str], nodename):
        self._sid = sid
        self.tid = tid
        self.endpoint_names = endpoint_names
        self.nodename = nodename
        self.client = Mock()
        self.client.set_state.side_effect = empty_future

    def get_id(self):
        return self._sid

    id = property(get_id)

    def get_client(self):
        return self.client

    async def expire_and_abort(self, timeout: float):
        # Only called on teardown
        pass


async def assert_agent_state(env_id: UUID, name: str, state: AgentStatus, sid: Optional[UUID]) -> None:
    agent = await data.Agent.get(env_id, name)

    assert agent.get_status() == state
    if state == AgentStatus.paused:
        assert agent.primary is None
        assert agent.paused
    elif state == AgentStatus.down:
        assert agent.primary is None
        assert not agent.paused
    elif state == AgentStatus.up:
        assert agent.primary is not None
        agent_instance = await data.AgentInstance.get_by_id(agent.primary)
        agent_proc = await data.AgentProcess.get_one(sid=agent_instance.process)
        assert agent_proc.sid == sid
        assert agent.get_status() == AgentStatus.up


async def assert_state_agents(
    env_id: UUID,
    s1: AgentStatus,
    s2: AgentStatus,
    s3: AgentStatus,
    sid1: Optional[UUID] = None,
    sid2: Optional[UUID] = None,
    sid3: Optional[UUID] = None,
) -> None:
    await assert_agent_state(env_id, "agent1", s1, sid1)
    await assert_agent_state(env_id, "agent2", s2, sid2)
    await assert_agent_state(env_id, "agent3", s3, sid3)


def assert_state_agents_retry(
    env_id: UUID,
    s1: AgentStatus,
    s2: AgentStatus,
    s3: AgentStatus,
    sid1: Optional[UUID] = None,
    sid2: Optional[UUID] = None,
    sid3: Optional[UUID] = None,
) -> typing.Callable[[], None]:
    async def func() -> bool:
        try:
            await assert_state_agents(env_id, s1, s2, s3, sid1, sid2, sid3)
        except AssertionError:
            return False
        return True

    return func


@pytest.mark.asyncio(timeout=30)
async def test_primary_selection(server, environment):
    env_id = UUID(environment)
    env = await data.Environment.get_by_id(env_id)
    am = server.get_slice(SLICE_AGENT_MANAGER)

    # second env to detect crosstalk
    project = data.Project(name="test_x")
    await project.insert()
    env2 = data.Environment(name="testenv_x", project=project.id)
    await env2.insert()

    await data.Agent(environment=env.id, name="agent1", paused=True).insert()
    await data.Agent(environment=env.id, name="agent2", paused=False).insert()
    await data.Agent(environment=env.id, name="agent3", paused=False).insert()

    # one session
    ts1 = MockSession(uuid4(), env.id, {"agent1", "agent2"}, "ts1")
    await am.new_session(ts1, set(ts1.endpoint_names))
    await am._session_listener_actions.join()
    assert len(am.sessions) == 1

    # cross talk session
    ts2 = MockSession(uuid4(), env2.id, {"agent1", "agent2"}, "ts2")
    await am.new_session(ts2, set(ts2.endpoint_names))
    await am._session_listener_actions.join()
    assert len(am.sessions) == 2

    ts1.get_client().set_state.assert_called_with("agent2", enabled=True)
    ts1.get_client().reset_mock()
    await retry_limited(
        assert_state_agents_retry(env.id, AgentStatus.paused, AgentStatus.up, AgentStatus.down, sid2=ts1.id), 10
    )

    # test is_primary
    assert am.is_primary(env, ts1.id, "agent2")
    assert not am.is_primary(env, ts1.id, "agent1")
    assert not am.is_primary(env, uuid4(), "agent2")

    # alive
    await am.seen(ts1, set(ts1.endpoint_names))
    await am._session_listener_actions.join()
    assert len(am.sessions) == 2
    await retry_limited(
        assert_state_agents_retry(env.id, AgentStatus.paused, AgentStatus.up, AgentStatus.down, sid2=ts1.id), 10
    )

    # second session
    ts2 = MockSession(uuid4(), env.id, {"agent3", "agent2"}, "ts2")
    await am.new_session(ts2, set(ts2.endpoint_names))
    await am._session_listener_actions.join()
    assert len(am.sessions) == 3
    ts2.get_client().set_state.assert_called_with("agent3", enabled=True)
    ts2.get_client().reset_mock()
    await retry_limited(
        assert_state_agents_retry(env.id, AgentStatus.paused, AgentStatus.up, AgentStatus.up, sid2=ts1.id, sid3=ts2.id), 10
    )

    # test is_primary
    assert not am.is_primary(env, ts2.id, "agent1")
    assert not am.is_primary(env, ts1.id, "agent1")
    assert am.is_primary(env, ts1.id, "agent2")
    assert am.is_primary(env, ts2.id, "agent3")

    # expire first
    await am.expire(ts1, set(ts1.endpoint_names))
    await am._session_listener_actions.join()
    assert len(am.sessions) == 2
    ts2.get_client().set_state.assert_called_with("agent2", enabled=True)
    ts2.get_client().reset_mock()
    await retry_limited(
        assert_state_agents_retry(env.id, AgentStatus.paused, AgentStatus.up, AgentStatus.up, sid2=ts2.id, sid3=ts2.id), 10
    )

    # expire second
    await am.expire(ts2, set(ts2.endpoint_names))
    await am._session_listener_actions.join()
    assert len(am.sessions) == 1
    await retry_limited(assert_state_agents_retry(env.id, AgentStatus.paused, AgentStatus.down, AgentStatus.down), 10)

    # test is_primary
    assert not am.is_primary(env, ts1.id, "agent2")
    assert not am.is_primary(env, ts2.id, "agent3")


@pytest.mark.asyncio(timeout=30)
async def test_api(init_dataclasses_and_load_schema):
    project = data.Project(name="test")
    await project.insert()

    env = data.Environment(name="testenv", project=project.id)
    await env.insert()
    env2 = data.Environment(name="testenv2", project=project.id)
    await env2.insert()
    env3 = data.Environment(name="testenv3", project=project.id)
    await env3.insert()
    # Exact replica of one to detect crosstalk
    env4 = data.Environment(name="testenv4", project=project.id)
    await env4.insert()
    await data.Agent(environment=env.id, name="agent1", paused=True).insert()
    await data.Agent(environment=env.id, name="agent2", paused=False).insert()
    await data.Agent(environment=env.id, name="agent3", paused=False).insert()
    await data.Agent(environment=env2.id, name="agent4", paused=False).insert()
    await data.Agent(environment=env4.id, name="agent1", paused=True).insert()
    await data.Agent(environment=env4.id, name="agent2", paused=False).insert()
    await data.Agent(environment=env4.id, name="agent3", paused=False).insert()

    server = Mock()
    futures = Collector()
    server.add_background_task.side_effect = futures
    am = AgentManager(server, False)
    am.add_background_task = futures
    am.running = True

    # one session
    ts1 = MockSession(uuid4(), env.id, {"agent1", "agent2"}, "ts1")
    await am._register_session(ts1, set(ts1.endpoint_names), datetime.datetime.now())
    # second session
    ts2 = MockSession(uuid4(), env.id, ["agent3", "agent2"], "ts2")
    await am._register_session(ts2, set(ts2.endpoint_names), datetime.datetime.now())
    # third session
    ts3 = MockSession(uuid4(), env3.id, ["agentx"], "ts3")
    await am._register_session(ts3, set(ts3.endpoint_names), datetime.datetime.now())
    # fourth session
    ts4 = MockSession(uuid4(), env4.id, {"agent1", "agent2"}, "ts4")
    await am._register_session(ts4, set(ts4.endpoint_names), datetime.datetime.now())

    await futures.proccess()
    assert len(am.sessions) == 4

    code, all_agents = await am.list_agent_processes(None, None)
    assert code == 200

    shouldbe = {
        "processes": [
            {
                "first_seen": UNKWN,
                "expired": None,
                "hostname": "ts1",
                "last_seen": UNKWN,
                "endpoints": [
                    {"id": UNKWN, "name": "agent1", "process": UNKWN},
                    {"id": UNKWN, "name": "agent2", "process": UNKWN},
                ],
                "environment": env.id,
            },
            {
                "first_seen": UNKWN,
                "expired": None,
                "hostname": "ts2",
                "last_seen": UNKWN,
                "endpoints": [
                    {"id": UNKWN, "name": "agent2", "process": UNKWN},
                    {"id": UNKWN, "name": "agent3", "process": UNKWN},
                ],
                "environment": env.id,
            },
            {
                "first_seen": UNKWN,
                "expired": None,
                "hostname": "ts3",
                "last_seen": UNKWN,
                "endpoints": [{"id": UNKWN, "name": "agentx", "process": UNKWN}],
                "environment": env3.id,
            },
            {
                "first_seen": UNKWN,
                "expired": None,
                "hostname": "ts4",
                "last_seen": UNKWN,
                "endpoints": [
                    {"id": UNKWN, "name": "agent1", "process": UNKWN},
                    {"id": UNKWN, "name": "agent2", "process": UNKWN},
                ],
                "environment": env4.id,
            },
        ]
    }

    assert_equal_ish(shouldbe, all_agents, sortby=["hostname", "name"])
    agentid = all_agents["processes"][0]["sid"]

    code, all_agents = await am.list_agent_processes(env.id, None)
    assert code == 200

    shouldbe = {
        "processes": [
            {
                "first_seen": UNKWN,
                "expired": None,
                "hostname": "ts1",
                "last_seen": UNKWN,
                "endpoints": [
                    {"id": UNKWN, "name": "agent1", "process": UNKWN},
                    {"id": UNKWN, "name": "agent2", "process": UNKWN},
                ],
                "environment": env.id,
            },
            {
                "first_seen": UNKWN,
                "expired": None,
                "hostname": "ts2",
                "last_seen": UNKWN,
                "endpoints": [
                    {"id": UNKWN, "name": "agent2", "process": UNKWN},
                    {"id": UNKWN, "name": "agent3", "process": UNKWN},
                ],
                "environment": env.id,
            },
        ]
    }

    assert_equal_ish(shouldbe, all_agents, sortby=["hostname", "name"])

    code, all_agents = await am.list_agent_processes(env2.id, None)
    assert code == 200

    shouldbe = {"processes": []}

    assert_equal_ish(shouldbe, all_agents)

    async def dummy_status():
        return Result(200, "X")

    ts1.get_client().get_status.side_effect = dummy_status
    report = await am.get_agent_process_report(agentid)
    assert (200, "X") == report

    report = await am.get_agent_process_report(uuid4())
    assert 404 == report[0]

    code, all_agents = await am.list_agents(None)
    assert code == 200
    shouldbe = {
        "agents": [
            {"name": "agent1", "paused": True, "last_failover": "", "primary": "", "environment": env.id, "state": "paused"},
            {"name": "agent2", "paused": False, "last_failover": UNKWN, "primary": UNKWN, "environment": env.id, "state": "up"},
            {"name": "agent3", "paused": False, "last_failover": UNKWN, "primary": UNKWN, "environment": env.id, "state": "up"},
            {"name": "agent1", "paused": True, "last_failover": "", "primary": "", "environment": env4.id, "state": "paused"},
            {
                "name": "agent2",
                "paused": False,
                "last_failover": UNKWN,
                "primary": UNKWN,
                "environment": env4.id,
                "state": "up",
            },
            {"name": "agent3", "paused": False, "last_failover": "", "primary": "", "environment": env4.id, "state": "down"},
            {"name": "agent4", "paused": False, "last_failover": "", "primary": "", "environment": env2.id, "state": "down"},
        ]
    }
    assert_equal_ish(shouldbe, all_agents, sortby=["environment", "name"])

    code, all_agents = await am.list_agents(env2)
    assert code == 200
    shouldbe = {
        "agents": [
            {"name": "agent4", "paused": False, "last_failover": "", "primary": "", "environment": env2.id, "state": "down"}
        ]
    }
    assert_equal_ish(shouldbe, all_agents)


@pytest.mark.asyncio(timeout=30)
async def test_expire_all_sessions_in_db(init_dataclasses_and_load_schema):
    project = data.Project(name="test")
    await project.insert()

    env = data.Environment(name="testenv", project=project.id)
    await env.insert()
    await data.Agent(environment=env.id, name="agent1", paused=True).insert()
    await data.Agent(environment=env.id, name="agent2", paused=False).insert()
    await data.Agent(environment=env.id, name="agent3", paused=False).insert()

    server = Mock()
    futures = Collector()
    server.add_background_task.side_effect = futures
    am = AgentManager(server, False)
    am.add_background_task = futures
    am.running = True

    # one session
    ts1 = MockSession(uuid4(), env.id, {"agent1", "agent2"}, "ts1")
    await am._register_session(ts1, set(ts1.endpoint_names), datetime.datetime.now())
    await futures.proccess()
    assert len(am.sessions) == 1
    ts1.get_client().set_state.assert_called_with("agent2", enabled=True)
    ts1.get_client().reset_mock()
    await assert_state_agents(env.id, AgentStatus.paused, AgentStatus.up, AgentStatus.down, sid2=ts1.id)

    # alive
    await am._seen_session(ts1, set(ts1.endpoint_names))
    await futures.proccess()
    assert len(am.sessions) == 1
    await assert_state_agents(env.id, AgentStatus.paused, AgentStatus.up, AgentStatus.down, sid2=ts1.id)

    # second session
    ts2 = MockSession(uuid4(), env.id, {"agent3", "agent2"}, "ts2")
    await am._register_session(ts2, set(ts2.endpoint_names), datetime.datetime.now())
    await futures.proccess()
    assert len(am.sessions) == 2
    ts2.get_client().set_state.assert_called_with("agent3", enabled=True)
    ts2.get_client().reset_mock()
    await assert_state_agents(env.id, AgentStatus.paused, AgentStatus.up, AgentStatus.up, sid2=ts1.id, sid3=ts2.id)

    # expire first
    await am._expire_session(ts1, set(ts1.endpoint_names), datetime.datetime.now())
    await futures.proccess()
    assert len(am.sessions) == 1
    ts2.get_client().set_state.assert_called_with("agent2", enabled=True)
    ts2.get_client().reset_mock()
    await assert_state_agents(env.id, AgentStatus.paused, AgentStatus.up, AgentStatus.up, sid2=ts2.id, sid3=ts2.id)

    # failover
    am = AgentManager(server, False)
    am.add_background_task = futures
    am.running = True
    await am._expire_all_sessions_in_db()

    # one session
    ts1 = MockSession(uuid4(), env.id, {"agent1", "agent2"}, "ts1")
    await am._register_session(ts1, set(ts1.endpoint_names), datetime.datetime.now())
    await futures.proccess()
    assert len(am.sessions) == 1
    ts1.get_client().set_state.assert_called_with("agent2", enabled=True)
    ts1.get_client().reset_mock()
    await assert_state_agents(env.id, AgentStatus.paused, AgentStatus.up, AgentStatus.down, sid2=ts1.id)

    # alive
    await am._seen_session(ts1, set(ts1.endpoint_names))
    await futures.proccess()
    assert len(am.sessions) == 1
    await assert_state_agents(env.id, AgentStatus.paused, AgentStatus.up, AgentStatus.down, sid2=ts1.id)

    # second session
    ts2 = MockSession(uuid4(), env.id, {"agent3", "agent2"}, "ts2")
    await am._register_session(ts2, set(ts2.endpoint_names), datetime.datetime.now())
    await futures.proccess()
    assert len(am.sessions) == 2
    ts2.get_client().set_state.assert_called_with("agent3", enabled=True)
    ts2.get_client().reset_mock()
    await assert_state_agents(env.id, AgentStatus.paused, AgentStatus.up, AgentStatus.up, sid2=ts1.id, sid3=ts2.id)

    # expire first
    await am._expire_session(ts1, set(ts1.endpoint_names), datetime.datetime.now())
    await futures.proccess()
    assert len(am.sessions) == 1
    ts2.get_client().set_state.assert_called_with("agent2", enabled=True)
    ts2.get_client().reset_mock()
    await assert_state_agents(env.id, AgentStatus.paused, AgentStatus.up, AgentStatus.up, sid2=ts2.id, sid3=ts2.id)

    # expire second
    await am._expire_session(ts2, set(ts2.endpoint_names), datetime.datetime.now())
    await futures.proccess()
    assert len(am.sessions) == 0
    await assert_state_agents(env.id, AgentStatus.paused, AgentStatus.down, AgentStatus.down)


async def assert_agent_db_state(
    tid: UUID, nr_procs: int, nr_non_expired_procs: int, nr_agent_instances: int, nr_non_expired_instances: int
) -> typing.Callable:
    """
    The database log is updated asynchronously. This method waits until
    the desired database state is reached.
    """

    async def is_db_state_reached():
        result = await data.AgentProcess.get_list(environment=tid)
        if len(result) != nr_procs:
            return False
        result = await data.AgentProcess.get_list(environment=tid, expired=None)
        if len(result) != nr_non_expired_procs:
            return False
        result = await data.AgentInstance.get_list(tid=tid)
        if len(result) != nr_agent_instances:
            return False
        result = await data.AgentInstance.get_list(tid=tid, expired=None)
        if len(result) != nr_non_expired_instances:
            return False
        return True

    await retry_limited(is_db_state_reached, 10)


@pytest.mark.asyncio
async def test_session_renewal(init_dataclasses_and_load_schema):
    """
    Agent got timeout but agent process was still running (e.g, network connectivity was disrupted).
    So the same agent process connects to the server with the same session id. This test verifies
    that the database state is updated correctly when an expired session is renewed.
    """
    project = data.Project(name="test")
    await project.insert()

    env = data.Environment(name="testenv", project=project.id)
    await env.insert()

    agent_manager = AgentManager()
    agent_manager._stopped = False
    agent_manager._stopping = False

    session_manager = SessionManager()
    sid = uuid4()
    tid = env.id
    endpoint = "vm1"
    session = Session(
        sessionstore=session_manager,
        sid=sid,
        hang_interval=1,
        timout=1,
        tid=tid,
        endpoint_names=[endpoint],
        nodename="test",
        disable_expire_check=True,
    )

    await assert_agent_db_state(tid, nr_procs=0, nr_non_expired_procs=0, nr_agent_instances=0, nr_non_expired_instances=0)
    await agent_manager._register_session(
        session=session, endpoint_names_snapshot=set(session.endpoint_names), now=datetime.datetime.now()
    )
    await assert_agent_db_state(tid, nr_procs=1, nr_non_expired_procs=1, nr_agent_instances=1, nr_non_expired_instances=1)
    await agent_manager._expire_session(
        session=session, endpoint_names_snapshot=set(session.endpoint_names), now=datetime.datetime.now()
    )
    await assert_agent_db_state(tid, nr_procs=1, nr_non_expired_procs=0, nr_agent_instances=1, nr_non_expired_instances=0)
    await agent_manager._register_session(
        session=session, endpoint_names_snapshot=set(session.endpoint_names), now=datetime.datetime.now()
    )
    await assert_agent_db_state(tid, nr_procs=1, nr_non_expired_procs=1, nr_agent_instances=1, nr_non_expired_instances=1)


@pytest.mark.asyncio
async def test_fix_corrupted_database(init_dataclasses_and_load_schema):
    """
    When the database connection is lost, the agent session information might get
    corrupted. This inconsistency should be fixed when a new session is registered.
    This test case verifies this behavior.
    """
    project = data.Project(name="test")
    await project.insert()

    env = data.Environment(name="testenv", project=project.id)
    await env.insert()

    agent_manager = AgentManager()
    agent_manager._stopped = False
    agent_manager._stopping = False

    session_manager = SessionManager()
    sid = uuid4()
    tid = env.id
    endpoint = "vm1"
    session = Session(
        sessionstore=session_manager,
        sid=sid,
        hang_interval=1,
        timout=1,
        tid=tid,
        endpoint_names=[endpoint],
        nodename="node1",
        disable_expire_check=True,
    )

    await assert_agent_db_state(tid, nr_procs=0, nr_non_expired_procs=0, nr_agent_instances=0, nr_non_expired_instances=0)
    await agent_manager._register_session(
        session=session, endpoint_names_snapshot=set(session.endpoint_names), now=datetime.datetime.now()
    )
    await assert_agent_db_state(tid, nr_procs=1, nr_non_expired_procs=1, nr_agent_instances=1, nr_non_expired_instances=1)
    await agent_manager._expire_session(
        session=session, endpoint_names_snapshot=set(session.endpoint_names), now=datetime.datetime.now()
    )
    await assert_agent_db_state(tid, nr_procs=1, nr_non_expired_procs=0, nr_agent_instances=1, nr_non_expired_instances=0)
    await agent_manager._register_session(
        session=session, endpoint_names_snapshot=set(session.endpoint_names), now=datetime.datetime.now()
    )
    await assert_agent_db_state(tid, nr_procs=1, nr_non_expired_procs=1, nr_agent_instances=1, nr_non_expired_instances=1)
    await agent_manager._expire_session(
        session=session, endpoint_names_snapshot=set(session.endpoint_names), now=datetime.datetime.now()
    )
    await assert_agent_db_state(tid, nr_procs=1, nr_non_expired_procs=0, nr_agent_instances=1, nr_non_expired_instances=0)

    # Make database corrupt
    instances = await data.AgentInstance.get_list(tid=tid)
    assert len(instances) == 1
    await instances[0].update(expired=None)
    # Assert corruption
    await assert_agent_db_state(tid, nr_procs=1, nr_non_expired_procs=0, nr_agent_instances=1, nr_non_expired_instances=1)

    # Session registration should fix the inconsistency
    await agent_manager._register_session(
        session=session, endpoint_names_snapshot=session.endpoint_names, now=datetime.datetime.now()
    )
    await assert_agent_db_state(tid, nr_procs=1, nr_non_expired_procs=1, nr_agent_instances=1, nr_non_expired_instances=1)


@pytest.mark.asyncio
async def test_session_creation_fails(server, environment, async_finalizer, caplog):
    """
    Verify that:
     * Session creation works correctly when the connectivity to the database works.
     * Session creation is refused when the connectivity to the database doesn't work.
       In that case the server state should stay consistent.
    """
    env_id = UUID(environment)
    agentmanager = server.get_slice(SLICE_AGENT_MANAGER)
    a = Agent(hostname="node1", environment=environment, agent_map={"agent1": "localhost"}, code_loader=False)
    await a.add_end_point_name("agent1")
    await a.start()
    async_finalizer(a.stop)

    # Wait until session is created
    await retry_limited(lambda: (env_id, "agent1") in agentmanager.tid_endpoint_to_session, 10)

    # Verify that the session is created correctly
    session = agentmanager.tid_endpoint_to_session[(env_id, "agent1")]
    session_manager = session._sessionstore
    assert len(agentmanager.sessions) == 1
    assert session in agentmanager.sessions.values()
    assert len(session_manager._sessions) == 1
    assert session in session_manager._sessions.values()
    assert "Heartbeat failed" not in caplog.text

    await a.stop()
    await session.expire(0)

    # Wait until session expired
    await retry_limited(lambda: len(agentmanager.tid_endpoint_to_session) == 0, 10)

    # Verify session expiry
    assert len(agentmanager.sessions) == 0
    assert len(agentmanager.tid_endpoint_to_session) == 0
    assert len(session_manager._sessions) == 0

    # Remove connectivity to the database
    await data.disconnect()
    caplog.clear()

    a = Agent(hostname="node1", environment=environment, agent_map={"agent1": "localhost"}, code_loader=False)
    await a.add_end_point_name("agent1")
    await a.start()

    # Verify that session creation fails and server state is stays consistent
    await retry_limited(lambda: "Heartbeat failed" in caplog.text, 10)
    assert len(agentmanager.sessions) == 0
    assert len(agentmanager.tid_endpoint_to_session) == 0
    assert len(session_manager._sessions) == 0


@pytest.mark.asyncio
async def test_agent_actions(server, client, async_finalizer):
    """
    Test the agent_action() and the all_agents_action() API call.
    """
    config.Config.set("config", "agent-deploy-interval", "0")
    config.Config.set("config", "agent-repair-interval", "0")
    agent_manager = server.get_slice(SLICE_AGENT_MANAGER)

    result = await client.create_project("test")
    assert result.code == 200
    project_id = result.result["project"]["id"]

    result = await client.create_environment(project_id=project_id, name="test1")
    env1_id = UUID(result.result["environment"]["id"])
    result = await client.create_environment(project_id=project_id, name="test2")
    env2_id = UUID(result.result["environment"]["id"])

    env_to_agent_map: Dict[UUID, agent.Agent] = {}

    async def start_agent(env_id: UUID, agent_names: List[str]) -> None:
        for agent_name in agent_names:
            await data.Agent(environment=env_id, name=agent_name, paused=False).insert()

        agent_map = {agent_name: "localhost" for agent_name in agent_names}
        a = agent.Agent(hostname="node1", environment=env_id, agent_map=agent_map, code_loader=False)
        for agent_name in agent_names:
            await a.add_end_point_name(agent_name)
        await a.start()
        async_finalizer(a.stop)
        env_to_agent_map[env_id] = a

    await start_agent(env1_id, ["agent1", "agent2"])
    await start_agent(env2_id, ["agent1"])

    await retry_limited(lambda: len(agent_manager.sessions) == 2, 10)

    async def assert_agents_paused(expected_statuses: Dict[Tuple[UUID, str], bool]) -> None:
        for (env_id, agent_name), paused in expected_statuses.items():
            # Check in-memory session state
            live_session_found = (env_id, agent_name) in agent_manager.tid_endpoint_to_session
            assert live_session_found != paused
            # Check database state
            agent_from_db = await data.Agent.get_one(environment=env_id, name=agent_name)
            assert agent_from_db.paused == paused
            assert (agent_from_db.primary is None) == paused
            if not paused:
                live_session = agent_manager.tid_endpoint_to_session[(env_id, agent_name)]
                agent_instance = await data.AgentInstance.get_by_id(agent_from_db.primary)
                assert agent_instance.process == live_session.id
            # Check agent state
            assert env_to_agent_map[env_id]._instances[agent_name].is_enabled() != paused

    await assert_agents_paused(
        expected_statuses={(env1_id, "agent1"): False, (env1_id, "agent2"): False, (env2_id, "agent1"): False}
    )
    # Pause agent1 and pause again
    for _ in range(2):
        result = await client.agent_action(tid=env1_id, name="agent1", action=AgentAction.pause.value)
        assert result.code == 200
        await assert_agents_paused(
            expected_statuses={(env1_id, "agent1"): True, (env1_id, "agent2"): False, (env2_id, "agent1"): False}
        )

    # Unpause agent1 and unpause again
    for _ in range(2):
        result = await client.agent_action(tid=env1_id, name="agent1", action=AgentAction.unpause.value)
        assert result.code == 200
        await assert_agents_paused(
            expected_statuses={(env1_id, "agent1"): False, (env1_id, "agent2"): False, (env2_id, "agent1"): False}
        )

    # Pause all agents in env1 and pause again
    for _ in range(2):
        result = await client.all_agents_action(tid=env1_id, action=AgentAction.pause.value)
        assert result.code == 200
        await assert_agents_paused(
            expected_statuses={(env1_id, "agent1"): True, (env1_id, "agent2"): True, (env2_id, "agent1"): False}
        )

    # Unpause all agents in env1 and unpause again
    for _ in range(2):
        result = await client.all_agents_action(tid=env1_id, action=AgentAction.unpause.value)
        assert result.code == 200
        await assert_agents_paused(
            expected_statuses={(env1_id, "agent1"): False, (env1_id, "agent2"): False, (env2_id, "agent1"): False}
        )

    # set up for halt test
    async def assert_agents_halt_state(env_id: UUID, agents_running: Dict[str, bool], halted: bool) -> None:
        """
        :param agents_running: dictionary of agents that were running before environment halting.
        """
        for agent_name, running in agents_running.items():
            db_agent: data.Agent = await data.Agent.get_one(environment=env_id, name=agent_name)
            assert db_agent.unpause_on_resume is (running if halted else None)

    result = await client.agent_action(tid=env1_id, name="agent1", action=AgentAction.pause.value)
    assert result.code == 200
    await assert_agents_paused(
        expected_statuses={(env1_id, "agent1"): True, (env1_id, "agent2"): False, (env2_id, "agent1"): False}
    )
    await assert_agents_halt_state(env1_id, {"agent1": False, "agent2": True}, False)
    await assert_agents_halt_state(env2_id, {"agent1": True}, False)

    # halt environment
    result = await client.halt_environment(env1_id)
    assert result.code == 200
    await assert_agents_paused(
        expected_statuses={(env1_id, "agent1"): True, (env1_id, "agent2"): True, (env2_id, "agent1"): False}
    )
    await assert_agents_halt_state(env1_id, {"agent1": False, "agent2": True}, True)
    await assert_agents_halt_state(env2_id, {"agent1": True}, False)

    # verify agent actions are not allowed in halted state
    result = await client.agent_action(tid=env1_id, name="agent1", action=AgentAction.unpause.value)
    assert result.code == 403
    assert "message" in result.result
    assert result.result["message"] == "Access denied: Can not pause or unpause agents when the environment has been halted."
    result = await client.agent_action(tid=env1_id, name="agent2", action=AgentAction.pause.value)
    assert result.code == 403
    result = await client.agent_action(tid=env2_id, name="agent1", action=AgentAction.pause.value)
    assert result.code == 200

    # resume environment, make sure only agents that were already running are unpaused
    result = await client.resume_environment(env1_id)
    assert result.code == 200
    await assert_agents_paused(
        expected_statuses={(env1_id, "agent1"): True, (env1_id, "agent2"): False, (env2_id, "agent1"): True}
    )
    await assert_agents_halt_state(env1_id, {"agent1": False, "agent2": True}, False)
    await assert_agents_halt_state(env2_id, {"agent1": False}, False)


@pytest.mark.asyncio
async def test_process_already_terminated(server, environment):
    """
    This test case tests whether the termination of autostarted agents (processes) happens correctly,
    when one of the processes has already terminated.
    """
    env_id = UUID(environment)
    env = await data.Environment.get_by_id(env_id)

    autostarted_agent_manager = server.get_slice(SLICE_AUTOSTARTED_AGENT_MANAGER)
    await autostarted_agent_manager._ensure_agents(env=env, agents=["internal"])
    assert len(autostarted_agent_manager._agent_procs) == 1

    # Terminate process
    autostarted_agent_manager._agent_procs[env_id].terminate()

    # This call shouldn't raise an exception
    await autostarted_agent_manager._terminate_agents()


@pytest.mark.asyncio
async def test_exception_occurs_while_processing_session_action(server, environment, async_finalizer, monkeypatch, caplog):
    """
    This test verifies that the consumer of the _session_listener_actions queue keeps working
    even when the an exception is thrown while handling an action.
    """
    agentmanager = server.get_slice(SLICE_AGENT_MANAGER)

    # Replace the _process_action() method with one that throws an excption
    old_process_action_function = agentmanager._process_action

    async def new_process_action_function(self, action: SessionAction, session: Session, timestamp: datetime):
        raise Exception("Failure")

    monkeypatch.setattr(agentmanager, "_process_action", new_process_action_function)

    assert len(agentmanager.sessions) == 0

    # Start agent
    a = Agent(hostname="node1", environment=environment, agent_map={"agent1": "localhost"}, code_loader=False)
    await a.add_end_point_name("agent1")
    await a.start()
    async_finalizer(a.stop)

    # Verify that an exception is thrown an no session is created
    await retry_limited(lambda: "An exception occurred while handling session action" in caplog.text, 10)
    assert len(agentmanager.sessions) == 0
    await a.stop()

    # Put original method in place.
    monkeypatch.setattr(agentmanager, "_process_action", old_process_action_function)

    # Start new agent
    a = Agent(hostname="node1", environment=environment, agent_map={"agent1": "localhost"}, code_loader=False)
    await a.add_end_point_name("agent1")
    await a.start()
    async_finalizer(a.stop)

    # Verify that session is created successfully -> Consumer didn't crash
    await retry_limited(lambda: len(agentmanager.sessions) == 1, 10)


@pytest.mark.asyncio
async def test_restart_on_environment_setting(server, client, environment, caplog):
    with caplog.at_level(logging.DEBUG):
        response = await client.environment_settings_set(tid=environment, id="autostart_agent_deploy_interval", value=300)
        assert response.code == 200

        def check_log_contains(caplog, loggerpart, level, msg):
            for record in caplog.get_records("call"):
                logger_name, log_level, message = record.name, record.levelno, record.message
                if msg in message and loggerpart in logger_name and level == log_level:
                    return True
            return False

        await retry_limited(
            lambda: check_log_contains(
                caplog, "inmanta.server.agentmanager", logging.DEBUG, "Restarting agents in environment"
            ),
            10,
        )


@pytest.mark.asyncio
async def test_failover_doesnt_make_paused_agent_primary(server, client, environment, agent_factory):
    """
    This test verifies that:

     * The standby agent doesn't take over when the agent is paused.
     * Session expiry doesn't make a standby agent primary when the agent is paused.
    """
    agent_name = "agent1"
    agent_config.agent_names.set(agent_name)
    agentmanager = server.get_slice(SLICE_AGENT_MANAGER)

    env_id = UUID(environment)
    env = await data.Environment.get_by_id(env_id)
    await agentmanager.ensure_agent_registered(env, agent_name)

    # Create two agent instances for the same agent (primary + standby agent)
    first_agent = await agent_factory(environment=environment, agent_map={agent_name: ""}, agent_names=[agent_name])
    # Wait until first_agent is elected as primary
    await retry_limited(lambda: len(agentmanager.tid_endpoint_to_session) == 1, timeout=10)
    await agent_factory(environment=environment, agent_map={agent_name: ""}, agent_names=[agent_name])

    # Wait until both agents have a session with the server
    await retry_limited(lambda: len(agentmanager.sessions) == 2, timeout=10)

    # Both agents have a session to the server, first_agent is primary
    assert len(agentmanager.sessions) == 2
    assert len(agentmanager.tid_endpoint_to_session) == 1
    # first_agent is primary
    assert agentmanager.tid_endpoint_to_session[(env_id, agent_name)].id == first_agent.sessionid

    # Pause agent1
    result = await client.agent_action(tid=environment, name=agent_name, action=AgentAction.pause.value)
    assert result.code == 200

    # Two agents have a session to the server, none of them is primary. The agent is paused
    assert len(agentmanager.sessions) == 2
    assert len(agentmanager.tid_endpoint_to_session) == 0

    # Stop one of the agents
    await first_agent.stop()

    # Wait until session expires
    await retry_limited(lambda: len(agentmanager.sessions) == 1 and len(agentmanager.tid_endpoint_to_session) == 0, timeout=10)

    # One agent has a session to the server, none of them is primary. The agent is paused
    assert len(agentmanager.sessions) == 1
    assert len(agentmanager.tid_endpoint_to_session) == 0


@pytest.mark.asyncio
async def test_add_internal_agent_when_missing_in_agent_map(server, environment, postgresql_client):
    """
    The internal agent should always be present in the autostart_agent_map. If it is not present, it is added when to
    auto-started agent is started. This test case verifies this behavior.
    """
    # Ensure agent1
    agentmanager = server.get_slice(SLICE_AGENT_MANAGER)
    env = await data.Environment.get_by_id(UUID(environment))
    await agentmanager.ensure_agent_registered(env=env, nodename="agent1")

    # Remove the internal agent from the autostart_agent_map
    query = "UPDATE public.environment SET settings=jsonb_set(settings, $1::text[], to_jsonb($2::jsonb), TRUE)"
    await postgresql_client.execute(query, [data.AUTOSTART_AGENT_MAP], "{}")

    # Assert internal agent not in autostart_agent_map
    env = await data.Environment.get_by_id(UUID(environment))
    autostart_agent_map = await env.get(data.AUTOSTART_AGENT_MAP)
    assert "internal" not in autostart_agent_map

    # Start autostarted agent1
    autostarted_agent_manager = server.get_slice(SLICE_AUTOSTARTED_AGENT_MANAGER)
    await autostarted_agent_manager._ensure_agents(env=env, agents=["agent1"])

    # Verify patch on autostart_agent_map
    env = await data.Environment.get_by_id(UUID(environment))
    autostart_agent_map = await env.get(data.AUTOSTART_AGENT_MAP)
    assert "internal" in autostart_agent_map


@pytest.mark.asyncio
async def test_error_handling_agent_fork(server, environment, monkeypatch):
    """
    Verifies resolution of issue: inmanta/inmanta-core#2777
    """
    exception_message = "The start of the agent failed"

    async def _dummy_fork_inmanta(
        self, args: List[str], outfile: Optional[str], errfile: Optional[str], cwd: Optional[str] = None
    ) -> subprocess.Process:
        raise Exception(exception_message)

    # Make the _fork_inmanta method raise an Exception
    monkeypatch.setattr(AutostartedAgentManager, "_fork_inmanta", _dummy_fork_inmanta)

    env = await data.Environment.get_by_id(UUID(environment))
    autostarted_agent_manager = server.get_slice(SLICE_AUTOSTARTED_AGENT_MANAGER)
    with pytest.raises(Exception) as excinfo:
        await autostarted_agent_manager._ensure_agents(env=env, agents=["internal"], restart=True)

    assert exception_message in str(excinfo.value)



"""
    Copyright 2017 Inmanta

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Contact: code@inmanta.com
"""

import pytest

from inmanta import resources
from inmanta.ast import ExternalException
from inmanta.resources import PARSE_ID_REGEX, PARSE_RVID_REGEX, Id, ResourceException, resource


class Base(resources.Resource):
    fields = ("a", "b", "c")

    @staticmethod
    def get_a(exporter, resource):
        return resource.a


@resources.resource("test_resource::Resource", agent="a", id_attribute="b")
class Resource(Base):
    fields = ("c", "d")
    map = {"d": lambda _, x: x.d}


def test_field_merge():
    assert len(Resource.fields) == 5


def test_fields_type():
    with pytest.raises(Exception):

        class Test(resources.Resource):
            fields = "z"


def test_fields_parent_type():
    with pytest.raises(Exception):

        class Base(resources.Resource):
            fields = "y"

        class Test(Base):
            fields = ("z",)


def test_resource_base(snippetcompiler):

    import inmanta.resources

    @resource("__config__::XResource", agent="agent", id_attribute="key")
    class MyResource(inmanta.resources.Resource):
        """
        A file on a filesystem
        """

        fields = ("key", "value", "agent")

    snippetcompiler.setup_for_snippet(
        """
        entity XResource:
            string key
            string agent
            string value
        end

        implement XResource using none

        implementation none for XResource:
        end

        XResource(key="key", agent="agent", value="value")
        """,
        autostd=False,
    )
    _version, json_value = snippetcompiler.do_export()

    assert len(json_value) == 1
    myresource = next(json_value.values().__iter__())

    assert myresource.key == "key"
    assert myresource.agent == "agent"
    assert myresource.value == "value"


def test_resource_base_with_method_key(snippetcompiler):

    import inmanta.resources

    @resource("__config__::XResource", agent="agent", id_attribute="key")
    class MyResource(inmanta.resources.Resource):
        """
        A file on a filesystem
        """

        fields = ("key", "value", "agent", "serialize")

        @staticmethod
        def get_serialize(_exporter, resource):
            return resource.key

    snippetcompiler.setup_for_snippet(
        """
        entity XResource:
            string key
            string agent
            string value
        end

        implement XResource using none

        implementation none for XResource:
        end

        XResource(key="key", agent="agent", value="value")
        """,
        autostd=False,
    )
    with pytest.raises(ResourceException):
        snippetcompiler.do_export()


def test_resource_with_keyword(snippetcompiler):

    import inmanta.resources

    @resource("__config__::YResource", agent="agent", id_attribute="key")
    class MyResource(inmanta.resources.Resource):
        """
        A file on a filesystem
        """

        fields = ("key", "value", "agent", "model")

        @staticmethod
        def get_model(_exporter, resource):
            return resource.key

    snippetcompiler.setup_for_snippet(
        """
         entity YResource:
             string key
             string agent
             string value
         end

         implement YResource using none

         implementation none for YResource:
         end

         YResource(key="key", agent="agent", value="value")
         """,
        autostd=False,
    )

    with pytest.raises(ResourceException):
        snippetcompiler.do_export()


def test_resource_with_private_method(snippetcompiler):

    import inmanta.resources

    @resource("__config__::YResource", agent="agent", id_attribute="key")
    class MyResource(inmanta.resources.Resource):
        """
        A file on a filesystem
        """

        fields = ("__setattr__", "key", "value", "agent")

    snippetcompiler.setup_for_snippet(
        """
        entity YResource:
            string key
            string agent
            string value
        end

        implement YResource using none

        implementation none for YResource:
        end

        YResource(key="key", agent="agent", value="value")
        """,
        autostd=False,
    )

    with pytest.raises(ResourceException):
        snippetcompiler.do_export()


def test_object_to_id(snippetcompiler):
    import inmanta.resources

    @resource("__config__::MYResource", agent="agent", id_attribute="key")
    class MyResource(inmanta.resources.Resource):
        fields = ("key", "value", "agent")

    snippetcompiler.setup_for_snippet(
        """
        import tests
        entity MYResource:
            string key
            string agent
            string value
        end

        implement MYResource using std::none

        x = MYResource(key="key", agent="agent", value="value")
        std::print(tests::get_id(x))
        """
    )

    snippetcompiler.do_export()


def test_resource_invalid_agent_name_annotation(snippetcompiler):

    import inmanta.resources

    with pytest.raises(ResourceException):

        @resource("__config__::XResource", agent=42, id_attribute="key")
        class MyResource(inmanta.resources.Resource):
            """
            A file on a filesystem
            """

            fields = ("key", "value", "agent")


def test_resource_invalid_agent_name_attribute_type(snippetcompiler):
    import inmanta.resources

    @resource("__config__::MYResource", agent="agent", id_attribute="key")
    class MyResource(inmanta.resources.Resource):
        fields = ("key", "value", "agent")

    snippetcompiler.setup_for_snippet(
        """
        import tests
        entity MYResource:
            string key
            int agent
            string value
        end

        implement MYResource using std::none

        x = MYResource(key="key", agent=47, value="value")
        std::print(tests::get_id(x))
        """
    )
    with pytest.raises(ExternalException):
        snippetcompiler.do_export()


def test_resource_invalid_agent_name_entity(snippetcompiler):
    import inmanta.resources

    @resource("__config__::MYResource", agent="agent", id_attribute="key")
    class MyResource(inmanta.resources.Resource):
        fields = ("key", "value", "agent")

    snippetcompiler.setup_for_snippet(
        """
        import tests

        entity AgentResource:
        end

        entity MYResource:
            string key
            string value
        end
        MYResource.agent [1] -- AgentResource.myresource [1]

        implement MYResource using std::none
        implement AgentResource using std::none

        x = MYResource(key="key", agent=AgentResource(), value="value")
        std::print(tests::get_id(x))
        """
    )
    with pytest.raises(ExternalException):
        snippetcompiler.do_export()


def test_is_resource_version_id():
    """
    Test whether the is_resource_version_id() method of the Id class works correctly.
    """
    assert Id.is_resource_version_id("test::Resource[agent,key=id],v=3")
    assert Id.is_resource_version_id("test::mod::Resource[agent,key=id],v=3")
    assert not Id.is_resource_version_id("test::Resource[agent,key=id]")
    assert not Id.is_resource_version_id("test::mod::Resource[agent,key=id]")
    assert not Id.is_resource_version_id("test::Resource")


def test_parse_id_regex():
    result = PARSE_ID_REGEX.search("test::Resource[agent,key=id],v=3")
    assert result is not None
    assert result.group("id") == "test::Resource[agent,key=id]"
    assert result.group("type") == "test::Resource"
    assert result.group("ns") == "test"
    assert result.group("class") == "Resource"
    assert result.group("hostname") == "agent"
    assert result.group("attr") == "key"
    assert result.group("value") == "id"
    assert result.group("version") == "3"

    result = PARSE_ID_REGEX.search("test::submodule::Resource[agent,key=id],v=3")
    assert result is not None
    assert result.group("id") == "test::submodule::Resource[agent,key=id]"
    assert result.group("type") == "test::submodule::Resource"
    assert result.group("ns") == "test::submodule"
    assert result.group("class") == "Resource"
    assert result.group("hostname") == "agent"
    assert result.group("attr") == "key"
    assert result.group("value") == "id"
    assert result.group("version") == "3"

    result = PARSE_ID_REGEX.search("test::Resource[agent,key=id]")
    assert result is not None
    assert result.group("id") == "test::Resource[agent,key=id]"
    assert result.group("type") == "test::Resource"
    assert result.group("ns") == "test"
    assert result.group("class") == "Resource"
    assert result.group("hostname") == "agent"
    assert result.group("attr") == "key"
    assert result.group("value") == "id"
    assert result.group("version") is None

    result = PARSE_ID_REGEX.search("test::submodule::Resource[agent,key=id]")
    assert result is not None
    assert result.group("id") == "test::submodule::Resource[agent,key=id]"
    assert result.group("type") == "test::submodule::Resource"
    assert result.group("ns") == "test::submodule"
    assert result.group("class") == "Resource"
    assert result.group("hostname") == "agent"
    assert result.group("attr") == "key"
    assert result.group("value") == "id"
    assert result.group("version") is None


def test_parse_rvid_regex():
    result = PARSE_RVID_REGEX.search("test::Resource[agent,key=id],v=3")
    assert result is not None
    assert result.group("id") == "test::Resource[agent,key=id]"
    assert result.group("type") == "test::Resource"
    assert result.group("ns") == "test"
    assert result.group("class") == "Resource"
    assert result.group("hostname") == "agent"
    assert result.group("attr") == "key"
    assert result.group("value") == "id"
    assert result.group("version") == "3"

    result = PARSE_RVID_REGEX.search("test::submodule::Resource[agent,key=id],v=3")
    assert result is not None
    assert result.group("id") == "test::submodule::Resource[agent,key=id]"
    assert result.group("type") == "test::submodule::Resource"
    assert result.group("ns") == "test::submodule"
    assert result.group("class") == "Resource"
    assert result.group("hostname") == "agent"
    assert result.group("attr") == "key"
    assert result.group("value") == "id"
    assert result.group("version") == "3"

    result = PARSE_RVID_REGEX.search("test::Resource[agent,key=id]")
    assert result is None

    result = PARSE_RVID_REGEX.search("test::submodule::Resource[agent,key=id]")
    assert result is None
